# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import calendar
from dataclasses import asdict
from unittest.mock import patch
from urllib.parse import quote

import ldap3
import pytest
from faker import Faker

from ucsschool.kelvin.client import (
    InvalidRequest,
    NoObject,
    PasswordsHashes,
    Session,
    User,
    UserResource,
)

fake = Faker()


API_VERSION = "v1"
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_SCHOOL_RESOURCE = f"{URL_BASE}/{API_VERSION}/schools/"
URL_SCHOOL_COLLECTION = URL_SCHOOL_RESOURCE
URL_SCHOOL_OBJECT = f"{URL_SCHOOL_RESOURCE}{{name}}"
URL_USER_RESOURCE = f"{URL_BASE}/{API_VERSION}/users/"
URL_USER_COLLECTION = URL_USER_RESOURCE
URL_USER_OBJECT = f"{URL_USER_COLLECTION}{{name}}"


@pytest.mark.asyncio
async def test_search_no_name_arg(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_user,
):
    school = new_school(1)[0]
    user1 = await new_school_user(school=school.name)
    user2 = await new_school_user(school=school.name)

    async with Session(**kelvin_session_kwargs) as session:
        objs = [obj async for obj in UserResource(session=session).search(school=school.name)]

    assert objs, f"No Users in school {school.name!r} found."
    assert len(objs) >= 2
    assert user1.dn in [obj.dn for obj in objs]
    assert user2.dn in [obj.dn for obj in objs]
    for obj in objs:
        if obj.dn == user1.dn:
            compare_kelvin_obj_with_test_data(obj, **asdict(user1))
        elif obj.dn == user2.dn:
            compare_kelvin_obj_with_test_data(obj, **asdict(user2))


@pytest.mark.asyncio
async def test_search_partial_name_arg(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_user,
):
    school = new_school(1)[0]
    user = await new_school_user(school=school.name)
    name = user.name
    name_len = len(name)
    name_begin = name[: int(name_len / 2)]
    name_end = name[len(name_begin) :]

    async with Session(**kelvin_session_kwargs) as session:
        objs1 = [
            obj
            async for obj in UserResource(session=session).search(
                school=school.name, name=f"{name_begin}*"
            )
        ]
        objs2 = [
            obj
            async for obj in UserResource(session=session).search(
                school=school.name, name=f"*{name_end}"
            )
        ]
    assert objs1, f"No User for school={school.name!r} and name='{name_begin}*' found."
    assert len(objs1) == 1
    assert user.dn == objs1[0].dn
    assert objs2, f"No User for school={school.name!r} and name='*{name_end}' found."
    assert len(objs2) == 1
    assert user.dn == objs2[0].dn


@pytest.mark.asyncio
async def test_search_inexact_school(new_school, kelvin_session_kwargs):
    school = new_school(1)[0]
    school_name = school.name
    school_name_len = len(school_name)
    school_name_begin = school_name[: int(school_name_len / 2)]

    async with Session(**kelvin_session_kwargs) as session:
        with pytest.raises(InvalidRequest):
            assert [
                obj
                async for obj in UserResource(session=session).search(
                    school=f"{school_name_begin}*"
                )
            ]


@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["birthday", "disabled", "expiration_date", "roles"])
async def test_search_exact(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_user,
    attr,
):
    user = await new_school_user()
    value = getattr(user, attr)

    async with Session(**kelvin_session_kwargs) as session:
        objs = [
            obj
            async for obj in UserResource(session=session).search(
                school=user.school, **{attr: value}
            )
        ]
    assert objs, f"No User for school={user.school!r} and {attr!r}={value!r} found."
    if attr in ("roles", "disabled"):
        assert len(objs) >= 1
        assert user.dn in [o.dn for o in objs]
    else:
        assert len(objs) == 1
        assert user.dn == objs[0].dn


@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["firstname", "lastname", "record_uid", "source_uid"])
async def test_search_inexact(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_user,
    attr,
):
    user = await new_school_user()
    value = getattr(user, attr)
    value_len = len(value)
    value_begin = value[: int(value_len / 2)]
    value_end = value[len(value_begin) :]

    async with Session(**kelvin_session_kwargs) as session:
        objs1 = [
            obj
            async for obj in UserResource(session=session).search(
                school=user.school, **{attr: f"{value_begin}*"}
            )
        ]
        objs2 = [
            obj
            async for obj in UserResource(session=session).search(
                school=user.school, **{attr: f"*{value_end}"}
            )
        ]
    assert objs1, f"No User for school={user.school!r} and {attr!r}='{value_begin}*' found."
    assert objs2, f"No User for school={user.school!r} and {attr!r}='*{value_end}' found."
    for obj in objs1:
        assert getattr(obj, attr).startswith(value_begin)
    for obj in objs2:
        assert getattr(obj, attr).endswith(value_end)


@pytest.mark.asyncio
async def test_get_from_url(
    compare_kelvin_obj_with_test_data, kelvin_session_kwargs, new_school_user
):
    user = await new_school_user()
    url = URL_USER_OBJECT.format(
        host=kelvin_session_kwargs["host"], school=user.school, name=user.name
    )
    async with Session(**kelvin_session_kwargs) as session:
        obj: User = await UserResource(session=session).get_from_url(url)
    compare_kelvin_obj_with_test_data(obj, **asdict(user))


@pytest.mark.asyncio
async def test_get(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, new_school_user):
    user = await new_school_user()
    async with Session(**kelvin_session_kwargs) as session:
        obj = await UserResource(session=session).get(name=user.name)
    compare_kelvin_obj_with_test_data(obj, **asdict(user))


@pytest.mark.asyncio
async def test_create(
    check_password,
    kelvin_session_kwargs,
    ldap_access,
    new_user_test_obj,
    schedule_delete_obj,
    new_workgroup,
):
    user_data = new_user_test_obj()
    wg_dn, wg_attr = await new_workgroup(school=user_data.school, users=[])
    user_data.workgroups = {wg_attr["school"]: [wg_attr["name"]]}

    async with Session(**kelvin_session_kwargs) as session:
        user_obj = User(session=session, **asdict(user_data))
        user_obj.udm_properties = {"title": fake.first_name()}
        schedule_delete_obj(object_type="user", name=user_data.name)
        await user_obj.save()
        print(f"Created new User: {user_obj.as_dict()!r}")

    assert f"ou={user_data.school}" in user_obj.dn
    ldap_filter = f"(&(uid={user_obj.name})(objectClass=ucsschoolType))"
    ldap_objs = await ldap_access.search(filter_s=ldap_filter)
    assert len(ldap_objs) == 1
    ldap_obj = ldap_objs[0]
    assert ldap_obj.entry_dn == user_obj.dn
    assert ldap_obj["uid"].value == user_obj.name
    assert ldap_obj["givenName"].value == user_obj.firstname
    assert ldap_obj["sn"].value == user_obj.lastname
    ldap_val_schools = ldap_obj["ucsschoolSchool"].value
    if isinstance(ldap_val_schools, str):
        ldap_val_schools = [ldap_val_schools]
    assert set(ldap_val_schools) == set(user_obj.schools)
    assert ldap_obj["univentionBirthday"].value == user_obj.birthday.strftime("%Y-%m-%d")
    if user_obj.disabled is True:
        exp_shadow_expire = 1
    elif user_obj.expiration_date:
        exp_shadow_expire = int(calendar.timegm(user_obj.expiration_date.timetuple()) / 3600 / 24)
    else:
        exp_shadow_expire = 0
    assert ldap_obj["shadowExpire"].value == exp_shadow_expire
    ldap_val_ucsschool_role = ldap_obj["ucsschoolRole"].value
    if isinstance(ldap_val_ucsschool_role, str):
        ldap_val_ucsschool_role = [ldap_val_ucsschool_role]
    assert set(ldap_val_ucsschool_role) == set(user_data.ucsschool_roles)
    assert ldap_obj["ucsschoolRecordUID"] == user_data.record_uid
    assert ldap_obj["ucsschoolSourceUID"] == user_data.source_uid
    await check_password(ldap_obj.entry_dn, user_data.password)
    assert ldap_obj["title"] == user_obj.udm_properties["title"]
    # check that user was added to workgroup
    wg_ldap_filter = f"(&(cn={wg_attr['school']}-{wg_attr['name']})(objectClass=ucsschoolGroup))"
    wg_ldap_objs = await ldap_access.search(filter_s=wg_ldap_filter)
    assert len(wg_ldap_objs) == 1
    wg_ldap_obj = wg_ldap_objs[0]
    assert user_obj.dn in wg_ldap_obj.uniqueMember, wg_ldap_obj


@pytest.mark.asyncio
@pytest.mark.parametrize("attr_value", [None, "", "omit"], ids=lambda x: repr(x))
@pytest.mark.parametrize(
    "attr_key,expected_message",
    [
        ("name", "No username was created"),
        ("record_uid", "source_uid or record_uid are not set"),
    ],
    ids=lambda x: repr(x),
)
async def test_create_user_raises_invalid_request_for_missing_username_and_record_uid(
    kelvin_session_kwargs,
    ldap_access,
    new_user_test_obj,
    attr_value,
    attr_key,
    expected_message,
):
    """
    Passing None, "" or leaving out the attribute should lead to an InvalidRequest on the
    server side. We have tests for the Kelvin server (test_create_without_username)
    which uses custom kelvin.json files.
    This test assumes that there is no schema defined on the Kelvin server.
    """
    user_data = new_user_test_obj()
    user_data_dict = asdict(user_data)
    user_data_dict[attr_key] = attr_value
    if attr_value == "omit":
        user_data_dict.pop(attr_key)
    async with Session(**kelvin_session_kwargs) as session:
        user_obj = User(session=session, **user_data_dict)
        with pytest.raises(InvalidRequest) as exc:
            await user_obj.save()
        assert expected_message in str(exc.value)


@pytest.mark.asyncio
@pytest.mark.parametrize("attr_value", [None, "", "omit"], ids=lambda x: repr(x))
async def test_create_user_email_allows_empty_non_required_attrs(
    kelvin_session_kwargs,
    ldap_access,
    new_user_test_obj,
    attr_value,
    schedule_delete_obj,
):
    """
    passing None, "" or leaving out the email attribute should __not__ lead to an InvalidRequest
    on the server side.
    This test assumes that there is no schema defined on the Kelvin server.
    """
    attr_key = "email"
    user_data = new_user_test_obj()
    user_data_dict = asdict(user_data)
    user_data_dict[attr_key] = attr_value
    if attr_value == "omit":
        user_data_dict.pop(attr_key)
    async with Session(**kelvin_session_kwargs) as session:
        user_obj = User(session=session, **user_data_dict)
        schedule_delete_obj(object_type="user", name=user_data.name)
        user = await user_obj.save()
        assert user.email is None


@pytest.mark.asyncio
async def test_create_with_password_hashes(
    check_password,
    kelvin_session_kwargs,
    password_hash,
    new_user_test_obj,
    schedule_delete_obj,
):
    user_data = new_user_test_obj()
    password, password_hashes = await password_hash()
    assert user_data.password != password
    user_data.password = None
    user_data.kelvin_password_hashes = PasswordsHashes(**asdict(password_hashes))

    async with Session(**kelvin_session_kwargs) as session:
        user_obj = User(session=session, **asdict(user_data))
        schedule_delete_obj(object_type="user", name=user_data.name)
        await user_obj.save()
        print(f"Created new User: {user_obj.as_dict()!r}")
    await check_password(user_obj.dn, password)


@pytest.mark.asyncio
async def test_modify_legal_wards(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
):
    user = await new_school_user(roles=["legal_guardian"])
    student = await new_school_user(roles=["student"])
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.legal_wards = [student.name]
        new_obj: User = await obj.save()
        assert new_obj.legal_wards == [student.name]
        obj: User = await user_resource.get(school=user.school, name=user.name)
        assert obj.legal_wards == [student.name]


@pytest.mark.asyncio
async def test_modify_legal_guardians(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
):
    user = await new_school_user(roles=["student"])
    legal_guardian = await new_school_user(roles=["legal_guardian"])
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.legal_guardians = [legal_guardian.name]
        new_obj: User = await obj.save()
        assert new_obj.legal_guardians == [legal_guardian.name]
        obj: User = await user_resource.get(school=user.school, name=user.name)
        assert obj.legal_guardians == [legal_guardian.name]


@pytest.mark.asyncio
async def test_modify_legal_wards_wrong_role(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
):
    user = await new_school_user(roles=["teacher"])
    student = await new_school_user(roles=["student"])
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.legal_wards = [student.name]
        with pytest.raises(InvalidRequest) as exc_info:
            await obj.save()
        assert (
            "Parameter 'legal_wards' only valid for user-role 'legal_guardian'"
            in exc_info.value.args[0]
        )


@pytest.mark.asyncio
async def test_modify_legal_guardians_wrong_role(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
):
    user = await new_school_user(roles=["teacher"])
    legal_guardian = await new_school_user(roles=["legal_guardian"])
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.legal_guardians = [legal_guardian.name]
        with pytest.raises(InvalidRequest) as exc_info:
            await obj.save()
        assert (
            "Parameter 'legal_guardians' only valid for user-role 'student'"
            in exc_info.value.args[0]
        )


@pytest.mark.asyncio
async def test_modify(
    check_password,
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
    new_user_test_obj,
    new_workgroup,
    ldap_access,
):
    user = await new_school_user()
    wg_dn, wg_attr = await new_workgroup(school=user.school, users=[])
    new_data = asdict(
        new_user_test_obj(
            name=user.name,
            roles=user.roles,
            school=user.school,
            schools=user.schools,
            ucsschool_roles=user.ucsschool_roles,
            udm_properties={"title": fake.first_name()},
            workgroups={wg_attr["school"]: [wg_attr["name"]]},
        )
    )
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        await check_password(obj.dn, user.password)
        assert new_data["firstname"] != obj.firstname
        for k, v in new_data.items():
            if k not in ("dn", "url"):
                setattr(obj, k, v)
        new_obj: User = await obj.save()
        assert new_obj is obj
        compare_kelvin_obj_with_test_data(new_obj, **obj.as_dict())
        # load fresh object
        fresh_obj: User = await user_resource.get(school=user.school, name=user.name)
        compare_kelvin_obj_with_test_data(fresh_obj, **new_obj.as_dict())
        compare_kelvin_obj_with_test_data(fresh_obj, **obj.as_dict())
        await check_password(fresh_obj.dn, new_data["password"])
        # check that user was added to workgroup
        wg_ldap_filter = (
            f"(&(cn={wg_attr['school']}-{wg_attr['name']})(objectClass=ucsschoolGroup))"
        )
        wg_ldap_objs = await ldap_access.search(filter_s=wg_ldap_filter)
        assert len(wg_ldap_objs) == 1
        wg_ldap_obj = wg_ldap_objs[0]
        assert obj.dn in wg_ldap_obj.uniqueMember


@pytest.mark.asyncio
async def test_modify_password_hashes(
    check_password,
    kelvin_session_kwargs,
    password_hash,
    new_school_user,
):
    user = await new_school_user()
    password, password_hashes = await password_hash()
    assert user.password != password
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=user.name)
        await check_password(obj.dn, user.password)
        obj.password = None
        obj.kelvin_password_hashes = PasswordsHashes(**asdict(password_hashes))
        await obj.save()
        fresh_obj: User = await user_resource.get(school=user.school, name=user.name)
        await check_password(fresh_obj.dn, password)


@pytest.mark.asyncio
async def test_move_change_name(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_user,
    schedule_delete_obj,
):
    user = await new_school_user()
    old_name = user.name
    new_name = fake.first_name()
    assert old_name != new_name
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=user.school, name=old_name)
        assert obj.name == old_name
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.name = new_name
        old_url = obj.url
        schedule_delete_obj(object_type="user", name=new_name)

        new_obj: User = await obj.save()
        assert new_obj is obj
        assert new_obj.name == new_name
        assert f"uid={new_name}" in new_obj.dn
        assert new_obj.url != old_url
        # load fresh object
        fresh_obj: User = await user_resource.get(school=user.school, name=new_name)
        assert fresh_obj.name == new_name
        assert fresh_obj.url != old_url
        compare_kelvin_obj_with_test_data(fresh_obj, **new_obj.as_dict())


@pytest.mark.asyncio
async def test_move_change_school(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_user,
    schedule_delete_obj,
):
    user = await new_school_user()
    old_school = user.school
    school1, school2 = new_school(2)
    ou1, ou2 = school1.name, school2.name
    new_school_ = ou1 if old_school == ou2 else ou2
    assert old_school != new_school_
    async with Session(**kelvin_session_kwargs) as session:
        user_resource = UserResource(session=session)
        obj: User = await user_resource.get(school=old_school, name=user.name)
        assert obj.school == old_school
        compare_kelvin_obj_with_test_data(obj, **asdict(user))
        obj.school = new_school_
        obj.schools = [new_school_]
        old_url = obj.url
        schedule_delete_obj(object_type="user", name=user.name)

        new_obj: User = await obj.save()
        assert new_obj is obj
        assert new_obj.name == user.name
        assert new_obj.school == new_school_
        assert f"ou={new_school_}" in new_obj.dn
        assert new_obj.url == old_url
        # load fresh object
        fresh_obj: User = await user_resource.get(school=new_school_, name=user.name)
        assert fresh_obj.name == user.name
        assert fresh_obj.school == new_school_
        assert fresh_obj.schools == [new_school_]
        assert fresh_obj.url == old_url
        compare_kelvin_obj_with_test_data(fresh_obj, **new_obj.as_dict())


@pytest.mark.asyncio
async def test_delete(kelvin_session_kwargs, new_school_user):
    user = await new_school_user()
    async with Session(**kelvin_session_kwargs) as session:
        obj = await UserResource(session=session).get(school=user.school, name=user.name)
        assert obj
        res = await obj.delete()
        assert res is None

    async with Session(**kelvin_session_kwargs) as session:
        with pytest.raises(NoObject):
            await UserResource(session=session).get(school=user.school, name=user.name)


@pytest.mark.asyncio
async def test_reload(
    kelvin_session_kwargs,
    ldap_access,
    new_school_user,
):
    user = await new_school_user()
    first_name_old = user.firstname
    first_name_new = fake.first_name()

    async with Session(**kelvin_session_kwargs) as session:
        obj: User = await UserResource(session=session).get(school=user.school, name=user.name)
        assert obj.firstname == first_name_old
        await obj.reload()
        assert obj.firstname == first_name_old
        await ldap_access.modify(obj.dn, {"givenName": [(ldap3.MODIFY_REPLACE, [first_name_new])]})
        await obj.reload()
        assert obj.firstname == first_name_new


@pytest.mark.asyncio
async def test_as_json(kelvin_session_kwargs, new_school_user):
    user = await new_school_user()
    async with Session(**kelvin_session_kwargs) as session:
        old_obj: User = await UserResource(session=session).get(school=user.school, name=user.name)
        old_obj.udm_properties["title"] = "before"

        new_user = User(**old_obj.as_dict(), session=session)
        new_user.udm_properties["title"] = "after"
        assert old_obj.udm_properties["title"] != new_user.udm_properties["title"]


@pytest.mark.asyncio
async def test_exists(kelvin_session_kwargs, new_school_user):
    user = await new_school_user()
    async with Session(**kelvin_session_kwargs) as session:
        assert await UserResource(session=session).exists(name=user.name)
        assert not await UserResource(session=session).exists(name="DOESNOTEXIST")
        with patch("ucsschool.kelvin.client.session.Session.head") as mock:
            f = asyncio.Future()
            f.set_result(True)
            mock.return_value = f
            await UserResource(session=session).exists(name=user.name)
            mock.assert_called_once()


@pytest.mark.parametrize("quoted", (True, False))
def test__from_kelvin_response_unquote(quoted: bool):
    expected_schools = ["☁", "☃"]
    url_schools = [quote(school) if quoted else school for school in expected_schools]
    test_response = {
        "name": "demo_student",
        "school": f"https://dummy.fqdn/ucsschool/kelvin/v1/schools/{url_schools[0]}",
        "firstname": "Demo",
        "lastname": "Student",
        "roles": ["https://dummy.fqdn/ucsschool/kelvin/v1/roles/student"],
        "schools": [
            f"https://dummy.fqdn/ucsschool/kelvin/v1/schools/{school}" for school in url_schools
        ],
        "birthday": "",
        "school_classes": {
            "DEMOSCHOOL": ["Democlass"],
        },
        "workgroups": {},
    }
    sc = User._from_kelvin_response(response=test_response)
    assert sc.schools == expected_schools
    assert sc.school == expected_schools[0]



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import contextlib
import copy
import sys
import warnings

import httpx
import pytest
from async_property import async_property

from ucsschool.kelvin.client import (
    Role,
    RoleResource,
    School,
    SchoolClass,
    SchoolClassResource,
    SchoolResource,
    User,
    UserResource,
    WorkGroup,
    WorkGroupResource,
)
from ucsschool.kelvin.client.session import BadSettingsWarning, Session

PY38 = sys.version_info >= (3, 8)

kelvin_session_kwargs_mock = {
    "username": "username",
    "password": "password",
    "host": "localhost",
    "verify": False,
}


class SessionMock:
    @async_property
    async def token(self) -> str:
        return "Token"


def call_kwargs(call_args):
    return call_args.kwargs if PY38 else call_args[1]


@pytest.mark.asyncio
async def test_session_closes_on_context_exit():
    async with Session(**kelvin_session_kwargs_mock) as session:
        assert session._client
    with pytest.raises(RuntimeError):
        print(session.client)


@pytest.mark.asyncio
@pytest.mark.parametrize("arg_client_tasks", range(-10, 11))
async def test_session_warn_max_client_tasks(arg_client_tasks):
    with warnings.catch_warnings(record=True) as w:
        async with Session(
            max_client_tasks=arg_client_tasks, **kelvin_session_kwargs_mock
        ) as session:
            assert session.max_client_tasks >= 4
        if arg_client_tasks < 4:
            assert len(w) == 1
            assert issubclass(w[-1].category, BadSettingsWarning)
            assert "max_client_tasks" in str(w[-1].message)


@pytest.mark.asyncio
async def test_session_default_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    async with Session(**kelvin_session_kwargs_mock) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com")
        assert call_kwargs(session.client.get.call_args)["timeout"] == 10.0


@pytest.mark.asyncio
async def test_session_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    kelvin_session_kwargs = dict(kelvin_session_kwargs_mock, timeout=5)
    async with Session(**kelvin_session_kwargs) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com")
        assert call_kwargs(session.client.get.call_args)["timeout"] == 5


@pytest.mark.asyncio
async def test_session_explicit_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    async with Session(**kelvin_session_kwargs_mock) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com", timeout=7)
        assert call_kwargs(session.client.get.call_args)["timeout"] == 7


@pytest.mark.asyncio
async def test_session_override_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = dict(kelvin_session_kwargs_mock, timeout=4)
    async with Session(**kelvin_session_kwargs) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com", timeout=6)
        assert call_kwargs(session.client.get.call_args)["timeout"] == 6


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cid",
    [(None, None), (None, "X-Foo"), ("12345", None), ("12345", "X-Foo")],
    ids=lambda x: repr(x),
)
async def test_correlation_id(cid, mocker):
    correlation_id, correlation_id_header = cid
    mocker.patch("httpx.AsyncClient.send", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = copy.deepcopy(kelvin_session_kwargs_mock)
    if correlation_id:
        kelvin_session_kwargs["request_id"] = correlation_id
    if correlation_id_header:
        kelvin_session_kwargs["request_id_header"] = correlation_id_header
    async with Session(**kelvin_session_kwargs) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com")
        async_client_send_call_args = httpx.AsyncClient.send.call_args
        headers = async_client_send_call_args[0][0].headers
        exp_header = correlation_id_header or "X-Request-ID"
        assert exp_header in headers
        assert headers[exp_header]
        if correlation_id:
            assert headers[exp_header] == correlation_id


@pytest.mark.asyncio
@pytest.mark.parametrize("language", [None, "de", "en"])
async def test_language_header(mocker, language):
    mocker.patch("httpx.AsyncClient.send", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = copy.deepcopy(kelvin_session_kwargs_mock)
    async with Session(**kelvin_session_kwargs, language=language) as session:
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com")
        async_client_send_call_args = httpx.AsyncClient.send.call_args
        headers = async_client_send_call_args[0][0].headers
        if language:
            assert headers.get("accept-language")
            assert headers["accept-language"] == language
        else:
            assert headers.get("accept-language") is None


@pytest.mark.asyncio
@pytest.mark.parametrize("language", [None, "de", "en"])
@pytest.mark.parametrize(
    "ObjectClass,ResourceClass",
    [
        (School, SchoolResource),
        (User, UserResource),
        (SchoolClass, SchoolClassResource),
        (WorkGroup, WorkGroupResource),
        (Role, RoleResource),
    ],
)
async def test_language_header_change_during_session(mocker, language, ObjectClass, ResourceClass):
    mocker.patch("httpx.AsyncClient.send", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = copy.deepcopy(kelvin_session_kwargs_mock)
    async with Session(**kelvin_session_kwargs, language="dummy_lang") as session:
        kelvin_obj: ObjectClass = ResourceClass(session=session, language=language)
        with contextlib.suppress(NotImplementedError):
            await session.get("http://example.com")
        async_client_send_call_args = httpx.AsyncClient.send.call_args
        headers = async_client_send_call_args[0][0].headers
        if language:
            assert headers.get("accept-language")
            assert headers["accept-language"] == language
            assert kelvin_obj.session.language == language
        else:
            assert headers.get("accept-language")
            assert headers["accept-language"] == "dummy_lang"
            assert kelvin_obj.session.language == "dummy_lang"



import base64
import copy
import datetime
import logging
import os
import random
import string
import subprocess
import sys
from dataclasses import asdict, dataclass, fields
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import allure
import factory
import faker
import httpx
import pytest
import urllib3
from docker.errors import NotFound as ContainerNotFound
from ldap3 import (
    ALL_ATTRIBUTES,
    AUTO_BIND_TLS_BEFORE_BIND,
    SIMPLE,
    SUBTREE,
    Connection,
    Entry,
    Server,
)
from ldap3.core.exceptions import LDAPBindError, LDAPExceptionError
from ldap3.utils.conv import escape_filter_chars
from ruamel.yaml import YAML
from urllib3.exceptions import InsecureRequestWarning

import docker
from ucsschool.kelvin.client import InvalidRequest, KelvinObject, NoObject, ServerError

API_VERSION = "v1"

CA_CERT_PATH: Optional[Path] = None
TEST_SERVER_YAML_FILENAME = Path(__file__).parent / "test_server.yaml"
UCS_LDAP_PORT = 7389
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_CLASS_RESOURCE = f"{URL_BASE}/{API_VERSION}/classes/"
URL_CLASS_COLLECTION = f"{URL_CLASS_RESOURCE}?school={{school}}"
URL_CLASS_OBJECT = f"{URL_CLASS_RESOURCE}{{school}}/{{name}}"
URL_ROLE_RESOURCE = f"{URL_BASE}/{API_VERSION}/roles/"
URL_ROLE_COLLECTION = URL_ROLE_RESOURCE
URL_ROLE_OBJECT = f"{URL_ROLE_RESOURCE}{{name}}"
URL_SCHOOL_RESOURCE = f"{URL_BASE}/{API_VERSION}/schools/"
URL_SCHOOL_COLLECTION = URL_SCHOOL_RESOURCE
URL_SCHOOL_OBJECT = f"{URL_SCHOOL_COLLECTION}{{name}}"
URL_USER_RESOURCE = f"{URL_BASE}/{API_VERSION}/users/"
URL_USER_COLLECTION = URL_USER_RESOURCE
URL_USER_OBJECT = f"{URL_USER_COLLECTION}{{name}}"
URL_WORKGROUP_RESOURCE = f"{URL_BASE}/{API_VERSION}/workgroups/"
URL_WORKGROUP_COLLECTION = f"{URL_WORKGROUP_RESOURCE}?school={{school}}"
URL_WORKGROUP_OBJECT = f"{URL_WORKGROUP_RESOURCE}{{school}}/{{name}}"
UDM_DOCKER_CONTAINER_NAME = "udm_rest_only"
KELVIN_DOCKER_CONTAINER_NAME = "kelvin-api"

fake = faker.Faker()
logger = logging.getLogger(__name__)
_handler = logging.StreamHandler()
_handler.setLevel(logging.DEBUG)
logger.addHandler(_handler)
logger.setLevel(logging.DEBUG)
urllib3.disable_warnings(category=InsecureRequestWarning)


class TestServerConfiguration(NamedTuple):
    host: str
    username: str
    user_dn: str
    password: str
    verify: Union[bool, str]


class BadTestServerConfig(Exception): ...


class ContainerIpUnknown(Exception): ...


class NoTestServerConfig(Exception): ...


class TestServerConnectionError(Exception):
    __test__ = False

    def __init__(self, msg: str = None, status: int = None, reason: str = None):
        self.reason = reason
        self.status = status
        msg = msg or reason
        super().__init__(msg)


class LDAPAccess:
    _base_dn: str = None

    def __init__(self, bind_dn: str, bind_pw: str, host: str, port: int = 7389):
        self.bind_dn = bind_dn
        self.bind_pw = bind_pw
        self.host = host
        self.port = port
        self.server = Server(host=self.host, port=self.port, get_info="ALL")

    @property
    def _connection_kwargs(self) -> Dict[str, Any]:
        return {
            "server": self.server,
            "user": self.bind_dn,
            "password": self.bind_pw,
            "auto_bind": AUTO_BIND_TLS_BEFORE_BIND,
            "authentication": SIMPLE,
            "read_only": False,
        }

    @property
    def base_dn(self) -> str:
        if not self._base_dn:
            with Connection(**self._connection_kwargs) as conn:
                res = [c for c in conn.server.info.naming_contexts if c != "cn=translog"]
                self.__class__._base_dn = res[0]
        return self._base_dn

    async def search(
        self,
        filter_s: str,
        attributes: List[str] = ALL_ATTRIBUTES,
        base: str = None,
        scope=SUBTREE,
        user: str = None,
        password: str = None,
    ) -> List[Entry]:
        """
        Search an LDAP directory.

        The `Entry` object has an attribute `entry_dn`, which is the DN of the object.
        The other attributes can be accessed as if it were a dict, but accessing through
        a `.value` attribute (see example below).
        `entry_attributes_as_dict` returns the all attributes as a single dict::

            results = ldap_access.search("uid=Admin*")
            for result in results:
                dn             = result.entry_dn
                uid            = result["uid"].value
                full_name      = result["displayName"].value
                all_attributes = result.entry_attributes_as_dict

        :param str filter_s: LDAP filter, must be surrounded by parenthesis
        :param list(str) attributes: list of attributes to retrieve. Special cases:
            `ldap3.ALL_ATTRIBUTES` ('*', default), `ldap3.NO_ATTRIBUTES` ('1.1'),
            `ldap3.ALL_OPERATIONAL_ATTRIBUTES` ('+')
        :param str base: DN where the search should start
        :param str scope: one of `ldap3.BASE`, `ldap3.LEVEL`, `ldap3.SUBTREE` (default)
        :param str user: DN to instead of the one given in `__init__()`
        :param str password: password to instead of the one given in `__init__()`
        :return: a list of ldap3.abstract.entry.Entry objects
        :rtype: list(ldap3.abstract.entry.Entry)
        """
        base = base or self.base_dn
        connection_kwargs = self._connection_kwargs.copy()
        if user:
            connection_kwargs["user"] = user
        if password:
            connection_kwargs["password"] = password
        try:
            with Connection(**connection_kwargs) as conn:
                conn.search(
                    search_base=base,
                    search_filter=filter_s,
                    attributes=attributes,
                    search_scope=scope,
                )
        except LDAPBindError as exc:  # pragma: no cover
            logger.error(
                "When connecting (binding) to %r with bind_dn %r: %s",
                self.server.host,
                connection_kwargs["user"],
                exc,
            )
            raise
        except LDAPExceptionError as exc:  # pragma: no cover
            logger.error(
                "When searching on %r with bind_dn %r (filter_s=%r attributes=%r "
                "base=%r scope=%r): %s",
                self.server.host,
                connection_kwargs["user"],
                filter_s,
                attributes,
                base,
                scope,
                exc,
            )
            raise
        return conn.entries

    async def modify(
        self,
        dn: str,
        changes: Dict[str, List[Tuple[str, List[str]]]],
        controls: Iterable[Tuple[str, bool, Any]] = None,
    ) -> Dict[str, Any]:
        """
        This wrapper around `Connection.modify()`. See
        https://ldap3.readthedocs.io/en/latest/modify.html

        - changes is a dictionary in the form
            {'attribute1': change), 'attribute2': [change, change, ...], ...}
        - change is (operation, [value1, value2, ...])
        - operation is 0 (MODIFY_ADD), 1 (MODIFY_DELETE), 2 (MODIFY_REPLACE),
            3 (MODIFY_INCREMENT)

        Example::

            await ldap_access.modify(
                dn,
                {"description": [(ldap3.MODIFY_REPLACE, ["something new"])]}
            )

        :param str dn: DN of object to modify
        :param dict changes: dictionary with changes
        :param list controls: see https://ldap3.readthedocs.io/en/latest/connection.html#controls
        :return:
        """
        try:
            with Connection(**self._connection_kwargs) as conn:
                conn.modify(dn=dn, changes=changes, controls=controls)
        except LDAPBindError as exc:  # pragma: no cover
            logger.error(
                "When connecting (binding) to %r with bind_dn %r: %s",
                self.server.host,
                self.bind_dn,
                exc,
            )
            raise
        except LDAPExceptionError as exc:  # pragma: no cover
            logger.error(
                "When modifying on %r with bind_dn %r (changes=%r): %s",
                self.server.host,
                self.bind_dn,
                changes,
                exc,
            )
            raise
        return conn.result

    async def get_dn_of_user(self, username: str) -> str:
        filter_s = f"(uid={escape_filter_chars(username)})"
        results = await self.search(filter_s, attributes=[])
        if len(results) == 1:
            return results[0].entry_dn
        elif len(results) > 1:
            raise RuntimeError(
                f"More than 1 result when searching LDAP with filter {filter_s!r}: {results!r}."
            )
        else:
            return ""


def _get_ip_of_container(container_name: str) -> str:
    docker_client = docker.from_env()
    container = docker_client.containers.get(container_name)
    for _, v in container.attrs["NetworkSettings"]["Networks"].items():
        try:
            return v["IPAddress"]
        except KeyError:  # pragma: no cover
            pass
    raise ContainerIpUnknown(
        f"Could not get IP address from container {container_name!r}."
    )  # pragma: no cover


def _start_stopped_container(container_name: str):
    docker_client = docker.from_env()
    container = docker_client.containers.get(container_name)
    if container.status != "running":  # pragma: no cover
        logger.info(
            f"Found stopped Docker container {container_name!r}. Trying to start and continue."
        )  # pragma: no cover
        container.start()


@pytest.fixture(scope="session")
def running_test_container():
    """
    :raises: ContainerIpUnknown
    :raises: ContainerNotFound
    """

    def _func() -> TestServerConfiguration:
        for container_name in (UDM_DOCKER_CONTAINER_NAME, KELVIN_DOCKER_CONTAINER_NAME):
            _start_stopped_container(container_name)
        ip = _get_ip_of_container(UDM_DOCKER_CONTAINER_NAME)
        server = TestServerConfiguration(
            host=ip,
            username="Administrator",
            user_dn="uid=Administrator,cn=users,dc=ucs-test,dc=intranet",
            password="univention",  # noqa: S106
            verify=False,
        )
        logger.info(
            f"Using Docker containers '{KELVIN_DOCKER_CONTAINER_NAME!r}' and "
            f"{UDM_DOCKER_CONTAINER_NAME}."
        )
        return server

    return _func


def tag_with_python_version(items):
    # Determine the version string once
    py_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    for item in items:
        # Dynamically add the label to every test result
        item.add_marker(allure.label("python_version", py_version))


def pytest_collection_modifyitems(items):
    tag_with_python_version(items)


@pytest.fixture(scope="session")
def ldap_credentials(test_server_configuration) -> Dict[str, Any]:
    return {
        "bind_dn": test_server_configuration.user_dn,
        "bind_pw": test_server_configuration.password,
        "host": test_server_configuration.host,
        "port": UCS_LDAP_PORT,
    }


@pytest.fixture(scope="session")
def ldap_access(ldap_credentials):
    return LDAPAccess(**ldap_credentials)


@pytest.fixture(scope="session")
def base_dn(ldap_access):
    return ldap_access.base_dn


@pytest.fixture(scope="session")
def load_test_server_yaml(ucs_ca_file_path):
    def _func(
        path: Union[str, Path] = TEST_SERVER_YAML_FILENAME,
    ) -> TestServerConfiguration:
        """
        :raises: FileNotFoundError
        :raises: TypeError
        """
        yaml = YAML(typ="rt")
        config = yaml.load(Path(path))
        verify = ucs_ca_file_path(config["host"])
        return TestServerConfiguration(verify=verify, **config)

    return _func


@pytest.fixture(scope="session")
def save_test_server_yaml():
    """
    This is here only to make things simpler for developers. It's not actually
    needed in the tests.
    """

    def _func(
        host: str,
        username: str,
        user_dn: str,
        password: str,
        path: Union[str, Path] = TEST_SERVER_YAML_FILENAME,
    ) -> None:
        """
        :raises: OSError (PermissionError etc)
        """
        yaml = YAML(typ="rt")
        yaml.indent = 4
        with open(path, "w") as fp:
            yaml.dump(
                {
                    "host": host,
                    "username": username,
                    "user_dn": user_dn,
                    "password": password,
                },
                fp,
            )

    return _func


def retrieve_kelvin_access_token(host: str, username: str, password: str, verify: bool) -> str:
    logger.info("Retrieving Kelvin access token...")
    try:
        resp = httpx.post(
            URL_TOKEN.format(host=host),
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={"username": username, "password": password},
            verify=verify,
        )
    except httpx.NetworkError as exc:
        raise TestServerConnectionError(
            f"Error retrieving token from Kelvin REST API: {exc}"
        ) from exc
    if resp.status_code != 200:
        raise TestServerConnectionError(  # pragma: no cover
            f"Error retrieving token from Kelvin REST API: "
            f"[{resp.status_code}] {resp.reason_phrase}",
            reason=resp.reason_phrase,
            status=resp.status_code,
        )
    json_resp = resp.json()
    return json_resp["access_token"]


@pytest.fixture(scope="session")
def kelvin_token(test_server_configuration) -> str:
    return retrieve_kelvin_access_token(
        host=test_server_configuration.host,
        username=test_server_configuration.username,
        password=test_server_configuration.password,
        verify=test_server_configuration.verify,
    )


@pytest.fixture(scope="session")
def json_headers(kelvin_token) -> Dict[str, str]:
    return {
        "accept": "application/json",
        "Authorization": f"Bearer {kelvin_token}",
        "Content-Type": "application/json",
    }


def _test_a_server_configuration(server_config: TestServerConfiguration) -> None:
    logger.info("Testing auth with Kelvin REST API...")
    retrieve_kelvin_access_token(
        host=server_config.host,
        username=server_config.username,
        password=server_config.password,
        verify=server_config.verify,
    )
    logger.info("OK: auth with Kelvin REST API.")


@pytest.fixture(scope="session")  # noqa: C901
def test_server_configuration(
    load_test_server_yaml, running_test_container
) -> TestServerConfiguration:
    """
    Get data of server used to run tests.

    :raises: BadTestServerConfig
    :raises: NoTestServerConfig
    """
    logger.info("Trying to load test server config from %r...", str(TEST_SERVER_YAML_FILENAME))
    try:
        server_configuration = load_test_server_yaml()
        _test_a_server_configuration(server_configuration)
    except FileNotFoundError:  # pragma: no cover
        logger.error("File not found: %r.", str(TEST_SERVER_YAML_FILENAME))
    except TypeError as exc:  # pragma: no cover
        raise BadTestServerConfig(f"Error in '{TEST_SERVER_YAML_FILENAME!s}': {exc!s}") from exc
    except TestServerConnectionError as exc:  # pragma: no cover
        pytest.exit(
            f"Error connecting to test server using credentials "
            f"from '{TEST_SERVER_YAML_FILENAME!s}': [{exc.status}] {exc.reason}.\n"
            f"Maybe remove/rename {TEST_SERVER_YAML_FILENAME.name!r} and try "
            f"the Docker solution?"
        )
    else:
        return server_configuration

    logger.info("Trying to use running Docker container %r...", KELVIN_DOCKER_CONTAINER_NAME)
    try:
        res = running_test_container()
        _test_a_server_configuration(res)
    except ContainerNotFound:
        logger.error("Container not found.")
    except ContainerIpUnknown as exc:
        raise BadTestServerConfig(str(exc)) from exc
    except TestServerConnectionError as exc:
        pytest.exit(
            f"Error connecting to test server using credentials for Docker "
            f"Docker container {KELVIN_DOCKER_CONTAINER_NAME!r}: [{exc.status}] {exc!s}"
        )
    else:
        return res

    raise NoTestServerConfig("No test server configuration found.")  # pragma: no cover


@pytest.fixture(scope="session")
def kelvin_session_kwargs(test_server_configuration) -> Dict[str, str]:
    return {
        "username": test_server_configuration.username,
        "password": test_server_configuration.password,
        "host": test_server_configuration.host,
        "verify": test_server_configuration.verify,
    }


@dataclass
class TestSchool:
    name: str
    display_name: str
    educational_servers: List[str] = None
    administrative_servers: List[str] = None
    class_share_file_server: str = None
    home_share_file_server: str = None
    ucsschool_roles: List[str] = None
    udm_properties: Dict[str, Any] = None
    dn: str = None
    url: str = None


class SchoolFactory(factory.Factory):
    class Meta:
        model = TestSchool

    name = factory.LazyFunction(lambda: f"testou{fake.pyint(1000, 9999)}")
    display_name = factory.Faker("text", max_nb_chars=50)
    educational_servers = factory.LazyAttribute(lambda o: [f"edu{o.name[:10]}"])
    administrative_servers = factory.LazyAttribute(lambda o: [f"adm{o.name[:10]}"])
    class_share_file_server = factory.LazyAttribute(lambda o: f"edu{o.name[:10]}")
    home_share_file_server = factory.LazyAttribute(lambda o: f"edu{o.name[:10]}")
    ucsschool_roles = factory.List([])
    udm_properties = factory.Dict({"description": None})
    dn = ""
    url = ""


@pytest.fixture
def new_school_test_obj() -> Callable[[], TestSchool]:
    return lambda: SchoolFactory()


@pytest.fixture
def demoschool_data(
    json_headers, kelvin_session_kwargs, test_server_configuration
) -> List[Dict[str, Any]]:
    response = httpx.get(
        URL_SCHOOL_COLLECTION.format(host=test_server_configuration.host),
        headers=json_headers,
        verify=test_server_configuration.verify,
    )
    json_resp = response.json()
    if not {"DEMOSCHOOL", "DEMOSCHOOL2"}.issubset({obj["name"] for obj in json_resp}):
        raise AssertionError(  # pragma: no cover
            "To run the tests properly you need to have two schools named "
            "'DEMOSCHOOL' and 'DEMOSCHOOL2' at the moment! Execute *on the "
            "host*: "
            "/usr/share/ucs-school-import/scripts/create_ou DEMOSCHOOL; "
            "/usr/share/ucs-school-import/scripts/create_ou DEMOSCHOOL2"
        )
    return json_resp


@pytest.fixture
def new_school(base_dn, demoschool_data):
    """Create a new OU (not really, simply returns DEMOSCHOOL and DEMOSCHOOL2)."""

    demoschool1_data = [dd for dd in demoschool_data if dd["name"] == "DEMOSCHOOL"][0]
    demoschool2_data = [dd for dd in demoschool_data if dd["name"] == "DEMOSCHOOL2"][0]

    def _create_random_schools(amount: int) -> List[TestSchool]:
        assert amount <= 2, "At the moment only one or two schools can be requested."
        demo_school = TestSchool(**demoschool1_data)
        demo_school_2 = TestSchool(**demoschool2_data)
        schools = [demo_school, demo_school_2]
        random.shuffle(schools)
        return schools[:amount]

    return _create_random_schools


@dataclass
class TestSchoolClass:
    name: str
    school: str
    description: str = None
    users: List[str] = None
    create_share: bool = True
    ucsschool_roles: List[str] = None
    udm_properties: Dict[str, Any] = None
    dn: str = None
    url: str = None


class SchoolClassFactory(factory.Factory):
    class Meta:
        model = TestSchoolClass

    name = factory.LazyFunction(lambda: f"test.{fake.user_name()}")
    school = "DEMOSCHOOL"
    description = factory.Faker("text", max_nb_chars=50)
    users = factory.List([])
    create_share = True
    ucsschool_roles = factory.List([])
    udm_properties = factory.Dict({"mailAddress": None})
    dn = ""
    url = ""


@pytest.fixture
def new_school_class_test_obj() -> Callable[[], TestSchoolClass]:
    return lambda: SchoolClassFactory()


@pytest.fixture
async def new_school_class(
    new_school,
    kelvin_session_kwargs,
    ldap_access,
    new_school_class_test_obj,
    http_request,
    schedule_delete_obj,
):
    """Create a new school class"""

    host = kelvin_session_kwargs["host"]
    collection_url = URL_CLASS_RESOURCE.format(host=host)

    async def _func(**kwargs) -> Tuple[str, Dict[str, Any]]:
        if "name" not in kwargs:
            kwargs["name"] = f"test.{fake.first_name()}"
        name = kwargs["name"]
        if "school" not in kwargs:
            test_school = new_school(1)[0]
            kwargs["school"] = test_school.name
        school = kwargs["school"]
        sc_data = new_school_class_test_obj()
        for k, v in kwargs.items():
            setattr(sc_data, k, v)
        data = asdict(sc_data)
        del data["dn"]
        del data["ucsschool_roles"]
        del data["url"]
        json_data = copy.deepcopy(data)
        json_data["school"] = URL_SCHOOL_OBJECT.format(host=host, name=school)
        json_data["users"] = [
            URL_USER_OBJECT.format(host=host, name=user_name) for user_name in json_data["users"]
        ]
        schedule_delete_obj(object_type="class", school=school, name=name)
        obj = http_request("post", url=collection_url, json=json_data)
        dn = obj["dn"]
        logger.info("Created new SchoolClass: %r", obj)

        dn0, _ = dn.split(",", 1)
        assert dn0 == f"cn={school}-{name}"
        ldap_objs = await ldap_access.search(filter_s=f"(&({dn0})(objectClass=ucsschoolGroup))")
        assert len(ldap_objs) == 1
        ldap_obj = ldap_objs[0]
        assert ldap_obj.entry_dn == dn
        assert ldap_obj["cn"].value == f"{school}-{name}"
        assert "ucsschoolGroup" in ldap_obj["objectClass"]
        assert ldap_obj["ucsschoolRole"].value == f"school_class:school:{school}"

        return dn, data

    yield _func


@dataclass
class TestWorkGroup:
    name: str
    school: str
    description: str = None
    users: List[str] = None
    email: str = None
    allowed_email_senders_users: List[str] = None
    allowed_email_senders_groups: List[str] = None
    create_share: bool = True
    ucsschool_roles: List[str] = None
    udm_properties: Dict[str, Any] = None
    dn: str = None
    url: str = None


class WorkGroupFactory(factory.Factory):
    class Meta:
        model = TestWorkGroup

    name = factory.LazyFunction(lambda: f"test.{fake.user_name()}")
    school = "DEMOSCHOOL"
    description = factory.Faker("text", max_nb_chars=50)
    users = factory.List([])
    email = None
    allowed_email_senders_users = []
    allowed_email_senders_groups = []
    create_share = True
    ucsschool_roles = factory.List([])
    udm_properties = factory.Dict({})
    dn = ""
    url = ""


@pytest.fixture
def new_workgroup_test_obj() -> Callable[[], TestWorkGroup]:
    return lambda: WorkGroupFactory()


@pytest.fixture
async def new_workgroup(
    new_school,
    kelvin_session_kwargs,
    ldap_access,
    new_workgroup_test_obj,
    http_request,
    schedule_delete_obj,
):
    """Create a new workgroup"""

    host = kelvin_session_kwargs["host"]
    collection_url = URL_WORKGROUP_RESOURCE.format(host=host)

    async def _func(**kwargs) -> Tuple[str, Dict[str, Any]]:
        if "name" not in kwargs:
            kwargs["name"] = f"test.{fake.first_name()}"
        name = kwargs["name"]
        if "school" not in kwargs:
            test_school = new_school(1)[0]
            kwargs["school"] = test_school.name
        school = kwargs["school"]
        sc_data = new_workgroup_test_obj()
        for k, v in kwargs.items():
            setattr(sc_data, k, v)
        data = asdict(sc_data)
        del data["dn"]
        del data["ucsschool_roles"]
        del data["url"]
        json_data = copy.deepcopy(data)
        json_data["school"] = URL_SCHOOL_OBJECT.format(host=host, name=school)
        json_data["users"] = [
            URL_USER_OBJECT.format(host=host, name=user_name) for user_name in json_data["users"]
        ]
        schedule_delete_obj(object_type="workgroup", school=school, name=name)
        obj = http_request("post", url=collection_url, json=json_data)
        dn = obj["dn"]
        logger.info("Created new workgroup: %r", obj)

        dn0, _ = dn.split(",", 1)
        assert dn0 == f"cn={school}-{name}"
        ldap_objs = await ldap_access.search(filter_s=f"(&({dn0})(objectClass=ucsschoolGroup))")
        assert len(ldap_objs) == 1
        ldap_obj = ldap_objs[0]
        assert ldap_obj.entry_dn == dn
        assert ldap_obj["cn"].value == f"{school}-{name}"
        assert "ucsschoolGroup" in ldap_obj["objectClass"]
        assert ldap_obj["ucsschoolRole"].value == f"workgroup:school:{school}"

        return dn, data

    yield _func


@dataclass
class TestUserPasswordsHashes:
    user_password: List[str]
    samba_nt_password: str
    krb_5_key: List[str]
    krb5_key_version_number: int
    samba_pwd_last_set: int


@dataclass
class TestUser:
    name: str
    school: str
    schools: List[str]
    firstname: str
    lastname: str
    birthday: datetime.date = None
    disabled: bool = False
    email: str = None
    expiration_date: datetime.date = None
    password: str = None
    record_uid: str = None
    roles: List[str] = None
    school_classes: Dict[str, List[str]] = None
    workgroups: Dict[str, List[str]] = None
    source_uid: str = None
    udm_properties: Dict[str, Any] = None
    ucsschool_roles: List[str] = None
    dn: str = None
    url: str = None
    kelvin_password_hashes: TestUserPasswordsHashes = None

    @classmethod
    def create(cls, **args):
        valid_fields = {f.name for f in fields(cls)}
        valid_args = {k: v for k, v in args.items() if k in valid_fields}
        return cls(**valid_args)


class UserFactory(factory.Factory):
    class Meta:
        model = TestUser

    name = factory.Faker("user_name")
    school = ""
    schools = factory.List([])
    firstname = factory.Faker("first_name")
    lastname = factory.Faker("last_name")
    birthday = factory.Faker("date_of_birth", minimum_age=6, maximum_age=18)
    disabled = False
    email = None
    expiration_date = factory.Faker("date_between", start_date="+1y", end_date="+10y")
    password = factory.Faker("password", length=20)
    record_uid = factory.LazyAttribute(lambda o: o.name)
    roles = factory.List([])
    school_classes = factory.Dict({})
    workgroups = factory.Dict({})
    source_uid = "TESTID"
    udm_properties = factory.Dict({"title": None})
    ucsschool_roles = factory.List([])
    dn = ""
    url = ""
    kelvin_password_hashes = None


@pytest.fixture  # noqa: C901
def new_user_test_obj(new_school):  # noqa: C901
    role_choices = (
        "staff",
        "student",
        "teacher",
        "teacher_and_staff",
        "school_admin",
        "legal_guardian",
    )

    def _func(**kwargs) -> TestUser:
        if "roles" not in kwargs:
            try:
                role = kwargs.pop("role")
            except KeyError:
                role = random.choice(role_choices)  # noqa: S311
            if role in (
                "staff",
                "student",
                "teacher",
                "school_admin",
                "legal_guardian",
            ):
                kwargs["roles"] = [role]
            elif role == "teacher_and_staff":
                kwargs["roles"] = ["staff", "teacher"]
            else:
                raise ValueError(  # pragma: no cover
                    f"Argument 'role' to new_user_test_obj() must be one of "
                    f"{', '.join(role_choices)}."
                )
        if "school" not in kwargs and "schools" not in kwargs:
            test_school = new_school(1)[0]
            kwargs["school"] = test_school.name
            kwargs["schools"] = [kwargs["school"]]
        if "school" not in kwargs:
            kwargs["school"] = sorted(kwargs["schools"])[0]  # pragma: no cover
        if "schools" not in kwargs:
            kwargs["schools"] = [kwargs["school"]]
        if "ucsschool_roles" not in kwargs:
            kwargs["ucsschool_roles"] = [
                f"{role}:school:{school}"
                for role in kwargs["roles"]
                for school in kwargs["schools"]
            ]
        user: TestUser = UserFactory(**kwargs)
        # ensure half names in test_user.test_search_inexact() are long enough
        if "firstname" not in kwargs and len(user.firstname) < 6:  # pragma: no cover
            user.firstname = f"{user.firstname}{fake.first_name()}"
        if "lastname" not in kwargs and len(user.lastname) < 6:  # pragma: no cover
            user.lastname = f"{user.lastname}{fake.last_name()}"
        user.name = user.name[:15]
        return user

    return _func


@pytest.fixture
def new_school_user(
    kelvin_session_kwargs,
    ldap_access,
    http_request,
    new_user_test_obj,
    schedule_delete_obj,
):
    """Create a new school user"""

    host = kelvin_session_kwargs["host"]
    collection_url = URL_USER_COLLECTION.format(host=host)

    async def _func(**kwargs) -> TestUser:  # Tuple[str, Dict[str, Any]]:
        user_obj: TestUser = new_user_test_obj(**kwargs)

        data = asdict(user_obj)

        del data["dn"]
        del data["ucsschool_roles"]
        del data["url"]
        json_data = copy.deepcopy(data)
        if json_data["birthday"]:
            json_data["birthday"] = json_data["birthday"].strftime("%Y-%m-%d")
        if json_data["expiration_date"]:
            json_data["expiration_date"] = json_data["expiration_date"].strftime("%Y-%m-%d")
        json_data["roles"] = [
            URL_ROLE_OBJECT.format(host=host, name=role) for role in json_data["roles"]
        ]
        json_data["school"] = URL_SCHOOL_OBJECT.format(host=host, name=json_data["school"])
        json_data["schools"] = [
            URL_SCHOOL_OBJECT.format(host=host, name=school) for school in json_data["schools"]
        ]
        schedule_delete_obj(object_type="user", name=user_obj.name)
        obj = http_request("post", url=collection_url, json=json_data)
        dn = obj["dn"]
        logger.info("Created new User, API response: %r", obj)
        dn0, _ = dn.split(",", 1)
        assert dn0 == f"uid={user_obj.name}"
        ldap_objs = await ldap_access.search(filter_s=f"(&({dn0})(objectClass=ucsschoolType))")
        assert len(ldap_objs) == 1
        ldap_obj = ldap_objs[0]
        assert ldap_obj.entry_dn == dn
        assert ldap_obj["uid"].value == user_obj.name
        assert "ucsschoolType" in ldap_obj["objectClass"]
        roles = [url.rsplit("/", 1)[-1] for url in user_obj.roles]
        ldap_val = ldap_obj["ucsschoolRole"].value
        if isinstance(ldap_val, str):
            ldap_val = {ldap_val}
        elif isinstance(ldap_val, Iterable):
            ldap_val = set(ldap_val)
        assert ldap_val == {
            f"{role}:school:{school.rsplit('/', 1)[-1]}"
            for role in roles
            for school in user_obj.schools
        }
        # role/school urls to names
        obj["school"] = obj["school"].rsplit("/", 1)[-1]
        for attr in ("roles", "schools"):
            obj[attr] = [url.rsplit("/", 1)[-1] for url in obj[attr]]
        obj["password"] = user_obj.password
        obj["birthday"] = datetime.datetime.strptime(obj["birthday"], "%Y-%m-%d").date()
        obj["expiration_date"] = datetime.datetime.strptime(
            obj["expiration_date"], "%Y-%m-%d"
        ).date()
        return TestUser.create(**obj)

    yield _func


@pytest.fixture
def schedule_delete_obj(http_request, json_headers, test_server_configuration):
    kelvin_objs: List[Tuple[str, Dict[str, Any]]] = []
    url_templates = {
        "class": URL_CLASS_OBJECT,
        "school": URL_SCHOOL_OBJECT,
        "user": URL_USER_OBJECT,
        "workgroup": URL_WORKGROUP_OBJECT,
    }

    def _func(object_type: str, **search_args) -> None:
        kelvin_objs.append((object_type, search_args))

    yield _func

    for kelvin_type, obj_search_args in kelvin_objs:
        logger.info("Deleting %r object with %r...", kelvin_type, obj_search_args)

        url_template = url_templates[kelvin_type]
        obj_url = url_template.format(host=test_server_configuration.host, **obj_search_args)
        try:
            http_request("delete", url=obj_url, return_json=False)
        except NoObject:
            logger.info(
                "Object does not exist (anymore): kelvin_type=%r obj_search_args=%r.",
                kelvin_type,
                obj_search_args,
            )
            continue


@pytest.fixture  # noqa: C901
def http_request(json_headers, kelvin_session_kwargs):  # noqa: C901
    def _func(
        http_method: str, url: str, return_json: bool = True, **kwargs
    ) -> Union[str, Dict[str, Any]]:
        if "headers" not in kwargs:
            kwargs["headers"] = json_headers
        if "verify" not in kwargs:
            kwargs["verify"] = kelvin_session_kwargs["verify"]
        if "timeout" not in kwargs:
            kwargs["timeout"] = 30.0
        req_method = getattr(httpx, http_method)
        response: httpx.Response = req_method(url, **kwargs)
        logger.debug(
            "%s %r (**%r) -> %r (%r)",
            http_method.upper(),
            url,
            kwargs,
            response.status_code,
            response.reason_phrase,
        )
        if 200 <= response.status_code <= 299:
            if return_json:
                return response.json()
            else:
                return response.text
        elif response.status_code == 404:
            raise NoObject(
                f"Object not found ({http_method.upper()} {url!r}).",
                reason=response.reason_phrase,
                status=response.status_code,
                url=url,
            )
        elif 400 <= response.status_code <= 499:  # pragma: no cover
            try:
                resp_json = response.json()
                if "detail" in resp_json:
                    resp_json = resp_json["detail"]
            except ValueError:
                resp_json = ""
            raise InvalidRequest(
                f"Kelvin REST API returned status {response.status_code}, reason "
                f"{response.reason_phrase!r}{f' ({resp_json})' if resp_json else ''} for "
                f"{http_method.upper()} {url!r}.",
                reason=response.reason_phrase,
                status=response.status_code,
                url=url,
            )
        else:
            raise ServerError(
                reason=response.reason_phrase, status=response.status_code, url=url
            )  # pragma: no cover

    return _func


@pytest.fixture(scope="session")
def ucs_ca_file_path():  # pragma: no cover
    global CA_CERT_PATH
    ucs_ca_ori_filename = "ucs-root-ca.crt"
    ip_chars = string.digits + "."

    def _func(host) -> Union[bool, str]:
        global CA_CERT_PATH
        if all(s in ip_chars for s in host):
            # need a hostname (not IP address) to verify SSL certificate
            return False
        resp = httpx.get(f"https://{host}/{ucs_ca_ori_filename}", verify=False)  # noqa: S501
        if resp.status_code != 200:
            return False
        CA_CERT_PATH = Path("/tmp/", f"{os.getpid()}_{host}_{ucs_ca_ori_filename}")  # noqa: S108
        with CA_CERT_PATH.open("w") as fp:
            fp.write(resp.text)
        return str(CA_CERT_PATH)

    yield _func
    if CA_CERT_PATH:
        CA_CERT_PATH.unlink()


@pytest.fixture
def compare_kelvin_obj_with_test_data(kelvin_session_kwargs):
    def _func(kelvin_obj: KelvinObject, **test_data):
        for test_data_attr, test_data_value in test_data.items():
            if test_data_attr == "password":
                # use check_password() to check this
                continue
            kelvin_obj_value = getattr(kelvin_obj, test_data_attr)
            if isinstance(test_data_value, list):
                assert set(kelvin_obj_value) == set(test_data_value)
            else:
                print(test_data_attr)
                assert kelvin_obj_value == test_data_value

    return _func


@pytest.fixture
def check_password(ldap_access):
    async def _func(bind_dn: str, bind_pw: str) -> None:
        search_kwargs = {
            "filter_s": f"({bind_dn.split(',')[0]})",
            "attributes": ["uid"],
            "user": bind_dn,
            "password": bind_pw,
        }
        logger.info("Testing login (making LDAP search) with: %r", search_kwargs)
        try:
            results = await ldap_access.search(**search_kwargs)
        except LDAPBindError as exc:
            raise AssertionError(
                f"Login fail with user={bind_dn!r} and password={bind_pw!r}: {exc!s}"
            ) from exc
        logger.info("Login success.")
        assert len(results) == 1
        result = results[0]
        expected_uid = bind_dn.split(",")[0].split("=")[1]
        assert expected_uid == result["uid"].value

    return _func


@pytest.fixture
def password_hash(check_password, ldap_access, new_school_user):
    async def _func(password: str = None) -> Tuple[str, TestUserPasswordsHashes]:
        password = password or fake.password(length=20)
        user = await new_school_user(password=password)
        await check_password(user.dn, user.password)
        # get hashes of user
        filter_s = f"(uid={user.name})"
        attributes = [
            "userPassword",
            "sambaNTPassword",
            "krb5Key",
            "krb5KeyVersionNumber",
            "sambaPwdLastSet",
        ]
        ldap_results = await ldap_access.search(filter_s=filter_s, attributes=attributes)
        if len(ldap_results) == 1:
            ldap_result = ldap_results[0]
        else:
            raise RuntimeError(
                f"More than 1 result when searching LDAP with filter "
                f"{filter_s!r}: {ldap_results!r}."
            )
        user_password = ldap_result["userPassword"].value
        if not isinstance(user_password, list):
            user_password = [user_password]
        user_password = [pw.decode("ascii") for pw in user_password]
        krb_5_key = [base64.b64encode(v).decode("ascii") for v in ldap_result["krb5Key"].value]
        return (
            password,
            TestUserPasswordsHashes(
                user_password=user_password,
                samba_nt_password=ldap_result["sambaNTPassword"].value,
                krb_5_key=krb_5_key,
                krb5_key_version_number=ldap_result["krb5KeyVersionNumber"].value,
                samba_pwd_last_set=ldap_result["sambaPwdLastSet"].value,
            ),
        )

    return _func


@pytest.fixture(scope="session")
def exec_with_ssh():
    def _func(cmd: List[str], host: str = None) -> int:
        if not Path("/usr/bin/ssh").exists() or not Path("/usr/bin/sshpass").exists():
            print(
                "'ssh' and/or 'sshpass' are not installed. "
                "Please install them to enable OU deletion."
            )
            return -1
        ssh_cmd = [
            "/usr/bin/sshpass",
            "-p",
            "univention",
            "/usr/bin/ssh",
            "-o",
            "StrictHostKeyChecking no",
            "-o",
            "UserKnownHostsFile /dev/null",
            f"root@{host}",
        ] + cmd
        print(f"ssh to {host!r} and execute: {cmd!r}...")
        return subprocess.call(ssh_cmd)  # noqa: S603

    return _func


@pytest.fixture(scope="session")
def delete_ou_using_ssh(base_dn, exec_with_ssh, kelvin_session_kwargs):
    host = kelvin_session_kwargs["host"]

    async def _func(ou_name: str):
        print(f"Deleting OU {ou_name!r} on host {host!r}...")
        dn = f"ou={ou_name},{base_dn}"
        exec_with_ssh(["/usr/sbin/udm", "container/ou", "remove", "--dn", dn], host)

        group_dns = [
            f"cn=admins-{ou_name.lower()},cn=ouadmins,cn=groups,{base_dn}",
            f"cn=OU{ou_name}-Klassenarbeit,cn=ucsschool,cn=groups,{base_dn}",
            f"cn=OU{ou_name.lower()}-DC-Edukativnetz,cn=ucsschool,cn=groups,{base_dn}",
            f"cn=OU{ou_name.lower()}-DC-Verwaltungsnetz,cn=ucsschool,cn=groups,{base_dn}",
            f"cn=OU{ou_name.lower()}-Member-Edukativnetz,cn=ucsschool,cn=groups,{base_dn}",
            f"cn=OU{ou_name.lower()}-Member-Verwaltungsnetz,cn=ucsschool,cn=groups,{base_dn}",
        ]
        group_dns_s = " ".join(f"'{dn}'" for dn in group_dns)
        cmd = f"'for DN in {group_dns_s}; do /usr/sbin/udm groups/group remove --dn \"$DN\"; done'"
        print(f"Deleting groups on host {host!r}: {group_dns!r}...")
        exec_with_ssh(["/bin/bash", "-c", cmd], host)

    return _func


@pytest.fixture
async def schedule_delete_ou_using_ssh(delete_ou_using_ssh):
    ous_created: List[str] = []

    def _func(ou_name: str):
        ous_created.append(ou_name)

    yield _func

    for ou_name in ous_created:
        await delete_ou_using_ssh(ou_name)


@pytest.fixture
async def mail_domain(ldap_access):
    mail_domain_objs = await ldap_access.search(
        filter_s="(univentionObjectType=mail/domain)", attributes=["cn"]
    )
    if not mail_domain_objs:
        raise AssertionError(  # pragma: no cover
            "To run the tests properly you need to have have at least one mail domain. "
            "Execute *on the host*: "
            'udm mail/domain create --position "cn=domain,cn=mail,$(ucr get ldap/base)" '
            '--set name="$(ucr get domainname)"'
        )
    return mail_domain_objs[0]["cn"].value



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
from dataclasses import asdict
from unittest.mock import patch
from urllib.parse import quote

import ldap3
import pytest
from faker import Faker

from ucsschool.kelvin.client import (
    InvalidRequest,
    NoObject,
    SchoolClass,
    SchoolClassResource,
    Session,
)

fake = Faker()


API_VERSION = "v1"
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_CLASS_RESOURCE = f"{URL_BASE}/{API_VERSION}/classes/"
URL_CLASS_COLLECTION = f"{URL_CLASS_RESOURCE}?school={{school}}"
URL_CLASS_OBJECT = f"{URL_CLASS_RESOURCE}{{school}}/{{name}}"
URL_SCHOOL_RESOURCE = f"{URL_BASE}/{API_VERSION}/schools/"
URL_SCHOOL_COLLECTION = URL_SCHOOL_RESOURCE
URL_SCHOOL_OBJECT = f"{URL_SCHOOL_COLLECTION}{{name}}"


@pytest.mark.asyncio
async def test_search_no_name_arg(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_class,
):
    school = new_school(1)[0]
    school_name = school.name
    sc1_dn, sc1_attr = await new_school_class(school=school_name)
    sc2_dn, sc2_attr = await new_school_class(school=school_name)

    async with Session(**kelvin_session_kwargs) as session:
        objs = [
            obj async for obj in SchoolClassResource(session=session).search(school=school_name)
        ]

    assert objs, f"No SchoolClass in school {school!r} found."
    assert len(objs) >= 2
    assert sc1_dn in [sc.dn for sc in objs]
    assert sc2_dn in [sc.dn for sc in objs]
    for obj in objs:
        if obj.dn == sc1_dn:
            compare_kelvin_obj_with_test_data(obj, **sc1_attr)
        if obj.dn == sc2_dn:
            compare_kelvin_obj_with_test_data(obj, **sc2_attr)


@pytest.mark.asyncio
async def test_search_partial_name_arg(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_class,
):
    school = new_school(1)[0]
    school_name = school.name
    # don't use usual name starting with 'test.', as leftovers of previous
    # tests will also match 'test.*'
    name = f"{fake.first_name()}{fake.first_name()}"
    sc1_dn, sc1_attr = await new_school_class(school=school_name, name=name)
    name_len = len(name)
    name_begin = name[: int(name_len / 2)]
    name_end = name[len(name_begin) :]

    async with Session(**kelvin_session_kwargs) as session:
        objs1 = [
            obj
            async for obj in SchoolClassResource(session=session).search(
                school=school_name, name=f"{name_begin}*"
            )
        ]
        objs2 = [
            obj
            async for obj in SchoolClassResource(session=session).search(
                school=school_name, name=f"*{name_end}"
            )
        ]
    assert objs1, f"No SchoolClass for school={school_name!r} and name='{name_begin}*' found."
    assert len(objs1) == 1
    assert sc1_dn == objs1[0].dn
    assert objs2, f"No SchoolClass for school={school_name!r} and name='*{name_end}' found."
    assert len(objs2) == 1
    assert sc1_dn == objs2[0].dn


@pytest.mark.asyncio
async def test_search_inexact_school(new_school, kelvin_session_kwargs):
    school = new_school(1)[0]
    school_name = school.name
    school_name_len = len(school_name)
    school_name_begin = school_name[: int(school_name_len / 2)]

    async with Session(**kelvin_session_kwargs) as session:
        with pytest.raises(InvalidRequest):
            assert [
                obj
                async for obj in SchoolClassResource(session=session).search(
                    school=f"{school_name_begin}*"
                )
            ]


@pytest.mark.asyncio
async def test_get_from_url(
    compare_kelvin_obj_with_test_data, kelvin_session_kwargs, new_school_class
):
    sc1_dn, sc1_attr = await new_school_class()
    school = sc1_attr["school"]
    name = sc1_attr["name"]
    url = URL_CLASS_OBJECT.format(host=kelvin_session_kwargs["host"], school=school, name=name)
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolClassResource(session=session).get_from_url(url)
    compare_kelvin_obj_with_test_data(obj, dn=sc1_dn, **sc1_attr)


@pytest.mark.asyncio
async def test_get_no_users(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school,
    new_school_class,
):
    sc_dn, sc_attr = await new_school_class()
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolClassResource(session=session).get(
            school=sc_attr["school"], name=sc_attr["name"]
        )
    compare_kelvin_obj_with_test_data(obj, dn=sc_dn, **sc_attr)


@pytest.mark.asyncio
async def test_get_with_users(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school,
    new_school_class,
    new_school_user,
):
    user1 = await new_school_user()
    user2 = await new_school_user(school=user1.school)
    sc_dn, sc_attr = await new_school_class(school=user1.school, users=[user1.name, user2.name])
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolClassResource(session=session).get(
            school=sc_attr["school"], name=sc_attr["name"]
        )
    assert set(obj.users) == {user1.name, user2.name}
    compare_kelvin_obj_with_test_data(obj, dn=sc_dn, **sc_attr)


@pytest.mark.parametrize("create_share", [True, False], ids=lambda x: f"create_share={x}")
@pytest.mark.asyncio
async def test_create(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    ldap_access,
    mail_domain,
    new_school_class_test_obj,
    schedule_delete_obj,
    new_school_user,
    test_server_configuration,
    create_share,
):
    sc_data = new_school_class_test_obj()
    user1 = await new_school_user(school=sc_data.school)
    user2 = await new_school_user(school=sc_data.school)
    sc_data.users = [user1.name, user2.name]
    sc_data.create_share = create_share
    async with Session(**kelvin_session_kwargs) as session:
        sc_kwargs = asdict(sc_data)
        sc_obj = SchoolClass(session=session, **sc_kwargs)
        sc_obj.udm_properties = {"mailAddress": f"{fake.first_name()}@{mail_domain}"}
        schedule_delete_obj(object_type="class", school=sc_data.school, name=sc_data.name)
        await sc_obj.save()
        print(f"Created new SchoolClass: {sc_obj.as_dict()!r}")

    ldap_filter = f"(&(cn={sc_data.school}-{sc_data.name})(objectClass=ucsschoolGroup))"
    ldap_objs = await ldap_access.search(filter_s=ldap_filter)
    assert len(ldap_objs) == 1
    ldap_obj = ldap_objs[0]
    assert ldap_obj.entry_dn == sc_obj.dn
    assert ldap_obj["cn"].value == f"{sc_data.school}-{sc_data.name}"
    assert ldap_obj["ucsschoolRole"].value == f"school_class:school:{sc_data.school}"
    assert ldap_obj["description"].value == sc_obj.description
    assert set(ldap_obj["uniqueMember"].value) == {user1.dn, user2.dn}
    assert ldap_obj["mailPrimaryAddress"] == sc_obj.udm_properties["mailAddress"]
    share_ldap_filter = f"(&(cn={sc_data.school}-{sc_data.name})(objectClass=ucsschoolShare))"
    share_ldap_objs = await ldap_access.search(filter_s=share_ldap_filter)
    if create_share:
        assert len(share_ldap_objs) == 1
        share_ldap_obj = share_ldap_objs[0]
        assert share_ldap_obj["cn"].value == f"{sc_data.school}-{sc_data.name}"
    else:
        assert len(share_ldap_objs) == 0


@pytest.mark.asyncio
async def test_modify(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    mail_domain,
    new_school_class,
    new_school_class_test_obj,
    test_server_configuration,
):
    sc1_dn, sc1_attr = await new_school_class(
        udm_properties={"mailAddress": f"{fake.first_name()}@{mail_domain}"}
    )
    new_data = asdict(new_school_class_test_obj())
    school = sc1_attr["school"]
    name = sc1_attr["name"]
    async with Session(**kelvin_session_kwargs) as session:
        sc_resource = SchoolClassResource(session=session)
        obj: SchoolClass = await sc_resource.get(school=school, name=name)
        compare_kelvin_obj_with_test_data(obj, dn=sc1_dn, **sc1_attr)
        for k, v in new_data.items():
            if k not in ("school", "name", "dn", "url", "ucsschool_roles"):
                setattr(obj, k, v)
        new_obj: SchoolClass = await obj.save()
        assert new_obj is obj
        assert new_obj.as_dict() == obj.as_dict()
        # load fresh object
        fresh_obj: SchoolClass = await sc_resource.get(school=school, name=name)
        assert fresh_obj.as_dict() == new_obj.as_dict()
        compare_kelvin_obj_with_test_data(fresh_obj, **obj.as_dict())


@pytest.mark.asyncio
async def test_move_change_name(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school_class,
    schedule_delete_obj,
):
    sc1_dn, sc1_attr = await new_school_class()
    school = sc1_attr["school"]
    old_name = sc1_attr["name"]
    new_name = fake.first_name()
    assert old_name != new_name
    async with Session(**kelvin_session_kwargs) as session:
        sc_resource = SchoolClassResource(session=session)
        obj: SchoolClass = await sc_resource.get(school=school, name=old_name)
        assert obj.name == old_name
        compare_kelvin_obj_with_test_data(obj, dn=sc1_dn, **sc1_attr)
        obj.name = new_name
        old_url = obj.url
        schedule_delete_obj(object_type="class", school=school, name=new_name)

        new_obj: SchoolClass = await obj.save()
        assert new_obj is obj
        assert new_obj.name == new_name
        assert new_obj.url != old_url
        # load fresh object
        fresh_obj: SchoolClass = await sc_resource.get(school=school, name=new_name)
        assert fresh_obj.name == new_name
        assert fresh_obj.url != old_url
        compare_kelvin_obj_with_test_data(fresh_obj, **new_obj.as_dict())


@pytest.mark.asyncio
async def test_move_change_school(
    compare_kelvin_obj_with_test_data,
    new_school,
    kelvin_session_kwargs,
    new_school_class,
    schedule_delete_obj,
):
    sc1_dn, sc1_attr = await new_school_class()
    old_school = sc1_attr["school"]
    school1, school2 = new_school(2)
    ou1, ou2 = school1.name, school2.name
    name = sc1_attr["name"]
    new_school_ = ou1 if old_school == ou2 else ou2
    assert old_school != new_school_
    async with Session(**kelvin_session_kwargs) as session:
        sc_resource = SchoolClassResource(session=session)
        obj: SchoolClass = await sc_resource.get(school=old_school, name=name)
        assert obj.school == old_school
        compare_kelvin_obj_with_test_data(obj, dn=sc1_dn, **sc1_attr)
        obj.school = new_school_
        schedule_delete_obj(object_type="class", school=new_school_, name=name)

        with pytest.raises(InvalidRequest) as exc_info:
            await obj.save()
        assert "Moving of class to other school is not allowed" in exc_info.value.args[0]


@pytest.mark.asyncio
async def test_delete(kelvin_session_kwargs, new_school_class):
    sc1_dn, sc1_attr = await new_school_class()
    school = sc1_attr["school"]
    name = sc1_attr["name"]
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolClassResource(session=session).get(school=school, name=name)
        assert obj
        res = await obj.delete()
        assert res is None

    async with Session(**kelvin_session_kwargs) as session:
        with pytest.raises(NoObject):
            await SchoolClassResource(session=session).get(school=school, name=name)


@pytest.mark.asyncio
async def test_reload(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    ldap_access,
    new_school_class,
):
    sc1_dn, sc1_attr = await new_school_class()
    school = sc1_attr["school"]
    name = sc1_attr["name"]
    description_old = sc1_attr["description"]
    description_new = fake.text(max_nb_chars=50)

    async with Session(**kelvin_session_kwargs) as session:
        obj: SchoolClass = await SchoolClassResource(session=session).get(school=school, name=name)
        assert obj.description == description_old
        await obj.reload()
        assert obj.description == description_old
        await ldap_access.modify(
            obj.dn, {"description": [(ldap3.MODIFY_REPLACE, [description_new])]}
        )
        await obj.reload()
        assert obj.description == description_new


@pytest.mark.asyncio
async def test_exists(kelvin_session_kwargs, new_school_class):
    sc1_dn, sc1_attr = await new_school_class()
    async with Session(**kelvin_session_kwargs) as session:
        assert await SchoolClassResource(session=session).exists(
            name=sc1_attr["name"], school=sc1_attr["school"]
        )
        assert not await SchoolClassResource(session=session).exists(
            name="DOESNOTEXIST", school=sc1_attr["school"]
        )
        with patch("ucsschool.kelvin.client.session.Session.head") as mock:
            f = asyncio.Future()
            f.set_result(True)
            mock.return_value = f
            await SchoolClassResource(session=session).exists(
                name=sc1_attr["name"], school=sc1_attr["school"]
            )
            mock.assert_called_once()


@pytest.mark.parametrize("quoted", (True, False))
def test__from_kelvin_response_unquote(quoted: bool):
    expected_school = "☁"
    url_school = quote(expected_school) if quoted else expected_school
    expected_users = ["☃", "☼", "AÇIKGÜN", "ÇAĞACAN"]
    url_users = [quote(user) if quoted else user for user in expected_users]
    test_response = {
        "name": "Democlass",
        "school": f"https://dummy.fqdn/ucsschool/kelvin/v1/schools/{url_school}",
        "users": [f"https://dummy.fqdn/ucsschool/kelvin/v1/users/{user}" for user in url_users],
    }
    sc = SchoolClass._from_kelvin_response(response=test_response)
    assert sc.users == expected_users
    assert sc.school == expected_school



"""Test connection and utility functions in `conftest` module."""

import datetime
import tempfile

import jwt
import pytest
from conftest import TestServerConnectionError, retrieve_kelvin_access_token
from faker import Faker
from ruamel.yaml import YAML

TOKEN_HASH_ALGORITHM = "HS256"  # noqa: S105
fake = Faker()


@pytest.fixture
def connection_data():
    def _func():
        return {
            # host=IP -> ucs_ca_file_path() will not try download of CA
            "host": f"{fake.pyint(min_value=10, max_value=250)}."
            f"{fake.pyint(min_value=10, max_value=250)}."
            f"{fake.pyint(min_value=10, max_value=250)}."
            f"{fake.pyint(min_value=10, max_value=250)}",
            "username": fake.first_name(),
            "user_dn": f"uid={fake.first_name()},cn=users,dc={fake.first_name()}",
            "password": fake.first_name(),
        }

    return _func


def test_load_test_server_yaml(load_test_server_yaml, connection_data):
    server = connection_data()
    yaml = YAML(typ="safe")
    with tempfile.NamedTemporaryFile(mode="w") as fp:
        yaml.dump(server, fp)
        fp.flush()
        config = load_test_server_yaml(fp.name)
        assert server == {
            "host": config.host,
            "username": config.username,
            "user_dn": config.user_dn,
            "password": config.password,
        }


def test_save_test_server_yaml(save_test_server_yaml, connection_data):
    server = connection_data()
    yaml = YAML(typ="safe")
    with tempfile.NamedTemporaryFile() as fp:
        save_test_server_yaml(**server, path=fp.name)
        fp.flush()
        fp.seek(0)
        config = yaml.load(fp)
        assert config == server


def test_retrieve_kelvin_access_token_fail(kelvin_session_kwargs):
    with pytest.raises(TestServerConnectionError):
        retrieve_kelvin_access_token(
            host=kelvin_session_kwargs["host"],
            username=fake.user_name(),
            password=fake.password(),
            verify=False,
        )


def test_retrieve_kelvin_access_token_success(kelvin_session_kwargs):
    token = retrieve_kelvin_access_token(**kelvin_session_kwargs)
    payload = jwt.decode(
        token, algorithms=[TOKEN_HASH_ALGORITHM], options={"verify_signature": False}
    )
    exp_payload_sub = {
        "username": kelvin_session_kwargs["username"],
        "kelvin_admin": True,
        "schools": [],
        "roles": [],
    }
    assert payload["sub"] == exp_payload_sub
    expiry = datetime.datetime.fromtimestamp(payload["exp"])
    assert expiry > datetime.datetime.now()



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import logging
from typing import Any, Dict, Iterable, List, Type
from urllib.parse import unquote

from .base import KelvinObject, KelvinResource
from .exceptions import InvalidRequest
from .session import Session

logger = logging.getLogger(__name__)


class WorkGroup(KelvinObject):
    _class_display_name = "Workgroup"
    _kelvin_attrs = KelvinObject._kelvin_attrs + [
        "school",
        "description",
        "users",
        "create_share",
    ]

    def __init__(
        self,
        name: str,
        school: str,
        *,
        description: str = None,
        users: List[str] = None,
        email: str = None,
        allowed_email_senders_users: List[str] = None,
        allowed_email_senders_groups: List[str] = None,
        create_share: bool = True,
        ucsschool_roles: List[str] = None,
        udm_properties: Dict[str, Any] = None,
        dn: str = None,
        url: str = None,
        session: Session = None,
        language: str = None,
    ):
        super().__init__(
            name=name,
            ucsschool_roles=ucsschool_roles,
            udm_properties=udm_properties,
            dn=dn,
            url=url,
            session=session,
            language=language,
        )
        self.school = school
        self.description = description
        self.users = users
        self.email = email
        self.allowed_email_senders_users = (
            [] if allowed_email_senders_users is None else allowed_email_senders_users
        )
        self.allowed_email_senders_groups = (
            [] if allowed_email_senders_groups is None else allowed_email_senders_groups
        )
        self.create_share = create_share
        self._resource_class = WorkGroupResource

    @classmethod
    def _from_kelvin_response(cls, response: Dict[str, Any]) -> "WorkGroup":
        # user urls to user names
        response["users"] = [unquote(url.rsplit("/", 1)[-1]) for url in response["users"]]
        # 'school' will be done in super class
        return super()._from_kelvin_response(response)

    def _to_kelvin_request_data(self) -> Dict[str, Any]:
        data = super()._to_kelvin_request_data()
        # users names to urls
        data["users"] = [f"{self.session.urls['user']}{user}" for user in data["users"]]
        return data


class WorkGroupResource(KelvinResource):
    class Meta:
        kelvin_object: Type[KelvinObject] = WorkGroup
        required_get_attrs: Iterable[str] = ("name", "school")
        required_save_attrs: Iterable[str] = ("name", "school")
        required_head_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ("school",)

    def __init__(self, session: Session, language: str = None):
        super().__init__(session=session, language=language)
        self.collection_url = self.session.urls["workgroup"]
        self.object_url = f"{self.session.urls['workgroup']}{{school}}/{{name}}"

    def _check_search_attrs(self, **kwargs) -> None:
        super()._check_search_attrs(**kwargs)
        if "*" in kwargs["school"]:
            raise InvalidRequest("Argument 'school' for searching workgroups must be exact.")



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import base64
import datetime
import logging
import warnings
from typing import Any, Dict, Iterable, List, Type, get_type_hints
from urllib.parse import unquote

from .base import KelvinObject, KelvinResource
from .exceptions import InvalidRequest
from .session import Session

logger = logging.getLogger(__name__)


class PasswordsHashes:
    user_password: List[str]
    samba_nt_password: str
    krb_5_key: List[str]
    krb5_key_version_number: int
    samba_pwd_last_set: int

    def __init__(
        self,
        user_password: List[str],
        samba_nt_password: str,
        krb_5_key: List[str],
        krb5_key_version_number: int,
        samba_pwd_last_set: int,
    ):
        self.user_password = user_password
        self.samba_nt_password = samba_nt_password
        self.krb_5_key = krb_5_key
        self.krb5_key_version_number = krb5_key_version_number
        self.samba_pwd_last_set = samba_pwd_last_set

    def as_dict(self) -> Dict[str, Any]:
        return dict((attr, getattr(self, attr)) for attr in get_type_hints(self).keys())

    def as_dict_with_ldap_attr_names(self) -> Dict[str, Any]:
        """
        Wrapper around `as_dict()` that renames the keys to those used in a UCS'
        OpenLDAP.
        """
        res = self.as_dict()
        res["userPassword"] = res.pop("user_password")
        res["sambaNTPassword"] = res.pop("samba_nt_password")
        res["krb5Key"] = res.pop("krb_5_key")
        res["krb5KeyVersionNumber"] = res.pop("krb5_key_version_number")
        res["sambaPwdLastSet"] = res.pop("samba_pwd_last_set")
        return res

    @property
    def krb_5_key_as_bytes(self) -> List[bytes]:
        """Value of `krb_5_key` as a list of bytes."""
        return [base64.b64decode(k) for k in self.krb_5_key]

    @krb_5_key_as_bytes.setter
    def krb_5_key_as_bytes(self, value: List[bytes]) -> None:
        """Set value of `krb_5_key` from a list of bytes."""
        if not isinstance(value, list):
            raise TypeError("Argument 'value' must be a list.")
        self.krb_5_key = [base64.b64encode(v).decode("ascii") for v in value]


class User(KelvinObject):
    _class_display_name = "User"
    _kelvin_attrs = KelvinObject._kelvin_attrs + [
        "school",
        "firstname",
        "lastname",
        "birthday",
        "disabled",
        "email",
        "expiration_date",
        "kelvin_password_hashes",
        "password",
        "record_uid",
        "roles",
        "schools",
        "school_classes",
        "workgroups",
        "source_uid",
        "legal_guardians",
        "legal_wards",
    ]

    def __init__(
        self,
        name: str = None,
        school: str = None,
        *,
        firstname: str = None,
        lastname: str = None,
        birthday: datetime.date = None,
        disabled: bool = False,
        email: str = None,
        expiration_date: datetime.date = None,
        password: str = None,
        record_uid: str = None,
        roles: List[str],
        schools: List[str],
        school_classes: Dict[str, List[str]] = None,
        workgroups: Dict[str, List[str]] = None,
        source_uid: str = None,
        udm_properties: Dict[str, Any] = None,
        ucsschool_roles: List[str] = None,
        kelvin_password_hashes: PasswordsHashes = None,
        dn: str = None,
        url: str = None,
        session: Session = None,
        language: str = None,
        legal_guardians: List[str] = None,
        legal_wards: List[str] = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            ucsschool_roles=ucsschool_roles,
            udm_properties=udm_properties,
            dn=dn,
            url=url,
            session=session,
            language=language,
        )
        self.school = school
        self.firstname = firstname
        self.lastname = lastname
        self.birthday = birthday
        self.disabled = disabled
        self.email = email
        self.expiration_date = expiration_date
        self.password = password
        self.record_uid = record_uid
        self.roles = roles
        self.schools = schools
        self.school_classes = school_classes or {}
        self.workgroups = workgroups or {}
        self.source_uid = source_uid
        self.kelvin_password_hashes = kelvin_password_hashes
        self._resource_class = UserResource
        self.legal_guardians = legal_guardians or []
        self.legal_wards = legal_wards or []

    @classmethod
    def _from_kelvin_response(cls, response: Dict[str, Any]) -> "User":
        for attr in ("roles", "schools", "legal_guardians", "legal_wards"):
            # turn urls to names ('school' will be done in super class)
            if attr in response:
                response[attr] = [unquote(url.rsplit("/", 1)[-1]) for url in response[attr]]
        if response["birthday"]:
            response["birthday"] = datetime.datetime.strptime(
                response["birthday"], "%Y-%m-%d"
            ).date()
        if "expiration_date" not in response:
            warnings.warn(
                "User attribute in Kelvin REST API response missing 'expiration_date' attribute. "
                "Server version probably < '1.5.1'.",
                RuntimeWarning,
                stacklevel=2,
            )
            response["expiration_date"] = None
        elif response["expiration_date"]:
            response["expiration_date"] = datetime.datetime.strptime(
                response["expiration_date"], "%Y-%m-%d"
            ).date()
        return super()._from_kelvin_response(response)

    def _to_kelvin_request_data(self) -> Dict[str, Any]:
        data = super()._to_kelvin_request_data()
        url_keys = {
            "roles": "role",
            "schools": "school",
            "legal_guardians": "user",
            "legal_wards": "user",
        }
        for attr in url_keys.keys():
            url_key = url_keys[attr]
            # role/school names to urls
            data[attr] = [f"{self.session.urls[url_key]}{value}" for value in data[attr]]
        if data["kelvin_password_hashes"]:
            data["kelvin_password_hashes"] = self.kelvin_password_hashes.as_dict()
        else:
            del data["kelvin_password_hashes"]
        if data["birthday"]:
            data["birthday"] = data["birthday"].strftime("%Y-%m-%d")
        if data["expiration_date"]:
            data["expiration_date"] = data["expiration_date"].strftime("%Y-%m-%d")
        return {key: value for key, value in data.items() if value is not None}


class UserResource(KelvinResource):
    class Meta:
        kelvin_object: Type[KelvinObject] = User
        required_get_attrs: Iterable[str] = ("name",)
        required_save_attrs: Iterable[str] = ("school", "roles")
        required_head_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ()

    def __init__(self, session: Session, language: str = None):
        super().__init__(session=session, language=language)
        self.collection_url = self.session.urls["user"]
        self.object_url = f"{self.session.urls['user']}{{name}}"

    def _check_search_attrs(self, **kwargs) -> None:
        super()._check_search_attrs(**kwargs)
        if "*" in kwargs.get("school", ""):
            raise InvalidRequest("Argument 'school' for searching users must be exact.")



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import pytest

from ucsschool.kelvin.client.role import Role
from ucsschool.kelvin.client.school import School
from ucsschool.kelvin.client.school_class import SchoolClass
from ucsschool.kelvin.client.user import User


@pytest.mark.parametrize("kelvin_object", [Role, School, SchoolClass, User])
def test_init_with_unknown_keyword_argument(kelvin_object):
    kelvin_object(
        name="test",
        school="test_school",
        schools=["test_school"],
        roles=[],
        notanargument=None,
    )



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.


class KelvinClientError(Exception):
    def __init__(self, msg: str = None, status: int = None, reason: str = None, url: str = None):
        self.reason = reason
        self.status = status
        self.url = url
        msg = msg or reason
        super().__init__(msg)


class InvalidRequest(KelvinClientError): ...


class InvalidToken(KelvinClientError): ...


class ServerError(KelvinClientError): ...


class NoObject(KelvinClientError): ...



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import datetime
import logging
import uuid
import warnings
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import httpx
import jwt
from async_property import async_property
from tenacity import (
    AsyncRetrying,
    RetryError,
    before_sleep_log,
    retry_if_exception_type,
    retry_if_result,
    stop_after_attempt,
    wait_exponential,
)

from .exceptions import InvalidRequest, InvalidToken, NoObject, ServerError

DN = str

API_VERSION = "v1"
TOKEN_HASH_ALGORITHM = "HS256"  # noqa: S105
TOKEN_LEEWAY = 30
SESSION_DEFAULT_RETRIES = 0
SESSION_DEFAULT_MIN_RETRY_PAUSE = 2  # seconds
SESSION_DEFAULT_MAX_RETRY_PAUSE = 20  # seconds
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_RESOURCE_CLASS = f"{URL_BASE}/{API_VERSION}/classes/"
URL_RESOURCE_ROLE = f"{URL_BASE}/{API_VERSION}/roles/"
URL_RESOURCE_SCHOOL = f"{URL_BASE}/{API_VERSION}/schools/"
URL_RESOURCE_USER = f"{URL_BASE}/{API_VERSION}/users/"
URL_RESOURCE_WORKGROUP = f"{URL_BASE}/{API_VERSION}/workgroups/"
logger = logging.getLogger(__name__)


class KelvinClientWarning(Warning): ...


class BadSettingsWarning(KelvinClientWarning): ...


@dataclass
class Token:
    expiry: datetime.datetime
    value: str

    @classmethod
    def from_str(cls, token_str: str) -> "Token":
        try:
            payload = jwt.decode(
                token_str,
                algorithms=[TOKEN_HASH_ALGORITHM],
                options={"verify_signature": False},
            )
        except jwt.PyJWTError as exc:
            raise InvalidToken(f"Error decoding token ({token_str!r}): {exc!s}") from exc
        if not isinstance(payload, dict) or "exp" not in payload:
            raise InvalidToken(
                f"Payload in token not a dict or missing 'exp' entry ({token_str!r})."
            )
        try:
            expiry = datetime.datetime.utcfromtimestamp(payload["exp"])
        except ValueError as exc:
            raise InvalidToken(f"Error parsing date in token ({token_str!r}): {exc!s}") from exc
        return cls(expiry=expiry, value=token_str)

    def is_valid(self) -> bool:
        if not self.expiry or not self.value:
            return False
        return datetime.datetime.utcnow() + datetime.timedelta(seconds=TOKEN_LEEWAY) <= self.expiry


class Session:
    def __init__(
        self,
        username: str,
        password: str,
        host: str,
        max_client_tasks: int = 10,
        request_id: str = None,
        request_id_header: str = "X-Request-ID",
        language: str = None,
        retries: int = SESSION_DEFAULT_RETRIES,
        **kwargs,
    ):
        if max_client_tasks < 4:
            txt = "Raising value of 'max_client_tasks' to its minimum of 4."
            warnings.warn(txt, BadSettingsWarning, stacklevel=2)
            logger.warning(txt)
            max_client_tasks = 4
        self.max_client_tasks = max_client_tasks
        self._client: Optional[httpx.AsyncClient] = None
        self._client_task_limiter = asyncio.Semaphore(max_client_tasks)
        self.username = username
        self.password = password
        self.host = host
        self.request_id = request_id or uuid.uuid4().hex
        self.request_id_header = request_id_header
        self.language = language
        self._retries = retries
        self._min_retry_pause = SESSION_DEFAULT_MIN_RETRY_PAUSE
        self._max_retry_pause = SESSION_DEFAULT_MAX_RETRY_PAUSE
        self.kwargs = kwargs
        self.urls = {
            "token": URL_TOKEN.format(host=host),
            "class": URL_RESOURCE_CLASS.format(host=host),
            "role": URL_RESOURCE_ROLE.format(host=host),
            "school": URL_RESOURCE_SCHOOL.format(host=host),
            "user": URL_RESOURCE_USER.format(host=host),
            "workgroup": URL_RESOURCE_WORKGROUP.format(host=host),
        }
        self._token: Optional[Token] = None

    async def __aenter__(self):
        self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def open(self) -> httpx.AsyncClient:
        if not self._client:
            self.kwargs["headers"] = self.kwargs.get("headers", {})
            self.kwargs["headers"]["Access-Control-Expose-Headers"] = self.request_id_header
            self.kwargs["headers"][self.request_id_header] = self.request_id
            self._client = httpx.AsyncClient(**self.kwargs)
        return self._client

    async def close(self) -> None:
        if self._client:
            await self._client.aclose()
        self._client = None

    @property
    def client(self) -> httpx.AsyncClient:
        if not self._client:
            raise RuntimeError("Session is closed.")
        return self._client

    @async_property
    async def token(self) -> str:
        if not self._token or not self._token.is_valid():
            resp_json = await self.post(
                self.urls["token"],
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={"username": self.username, "password": self.password},
            )
            self._token = Token.from_str(resp_json["access_token"])
        return self._token.value

    @async_property
    async def json_headers(self) -> Dict[str, str]:
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {await self.token}",
            "Content-Type": "application/json",
        }
        if self.language:
            headers["Accept-Language"] = self.language
        return headers

    async def request(
        self, async_request_method: Any, url: str, return_json: bool = True, **kwargs
    ) -> Union[str, int, Dict[str, Any]]:
        if "headers" not in kwargs:
            kwargs["headers"] = await self.json_headers
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.kwargs.get("timeout", 10.0)

        retrying = AsyncRetrying(
            stop=stop_after_attempt(self._retries + 1),
            wait=wait_exponential(
                multiplier=1, min=self._min_retry_pause, max=self._max_retry_pause
            ),
            retry=(
                retry_if_result(
                    lambda r: (
                        r.status_code
                        in (
                            httpx.codes.TOO_MANY_REQUESTS,
                            httpx.codes.BAD_GATEWAY,
                            httpx.codes.SERVICE_UNAVAILABLE,
                            httpx.codes.GATEWAY_TIMEOUT,
                        )
                    )
                )
                | retry_if_exception_type((httpx.RemoteProtocolError, httpx.NetworkError))
            ),
            before_sleep=before_sleep_log(logger, logging.WARNING),
            reraise=True,
        )

        try:
            response: httpx.Response = await retrying(async_request_method, url, **kwargs)
        except RetryError as exc:
            response = exc.last_attempt.result()

        try:
            resp_json = response.json()
            detail = resp_json["detail"] if "detail" in resp_json else ""
        except ValueError:
            detail = ""
            resp_json = {}

        if "Authorization" in kwargs["headers"]:
            kwargs["headers"]["Authorization"] = 10 * "*"
        if "data" in kwargs and "password" in kwargs["data"]:
            kwargs["data"]["password"] = 10 * "*"

        logger.debug(
            "[%s] %s %r (**%r) -> %r %r%s",
            self.request_id[:10],
            async_request_method.__name__.upper(),
            url,
            kwargs,
            response.status_code,
            response.reason_phrase,
            f" ({detail})" if detail else "",
        )
        if async_request_method == self.client.head:
            return response.status_code

        if 200 <= response.status_code <= 299:
            return resp_json if return_json else response.text
        elif response.status_code == 404:
            raise NoObject(
                f"Object not found ({async_request_method.__name__.upper()} {url!r}).",
                reason=detail if detail else response.reason_phrase,
                status=response.status_code,
                url=url,
            )
        elif 400 <= response.status_code <= 499:
            raise InvalidRequest(
                f"Kelvin REST API returned status {response.status_code}, reason "
                f"{response.reason_phrase!r}{f' ({detail})' if detail else ''} for "
                f"{async_request_method.__name__.upper()} {url!r}.",
                reason=detail if detail else response.reason_phrase,
                status=response.status_code,
                url=url,
            )
        else:
            raise ServerError(
                reason=response.reason_phrase, status=response.status_code, url=url
            )  # pragma: no cover

    async def delete(self, url: str, **kwargs) -> None:
        await self.request(self.client.delete, url, return_json=False, **kwargs)

    async def get(self, url: str, **kwargs) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        return await self.request(self.client.get, url, **kwargs)

    async def head(self, url: str, **kwargs) -> bool:
        return await self.request(self.client.head, url, **kwargs)

    # async def patch(self, url: str, **kwargs,) -> Dict[str, Any]:
    #     return await self.request(self.client.patch, url, **kwargs)

    async def post(self, url: str, **kwargs) -> Dict[str, Any]:
        return await self.request(self.client.post, url, **kwargs)

    async def put(self, url: str, **kwargs) -> Dict[str, Any]:
        return await self.request(self.client.put, url, **kwargs)



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.
import copy
import logging
from typing import Any, AsyncIterator, Dict, Iterable, List, TypeVar
from urllib.parse import unquote

from .exceptions import InvalidRequest, NoObject
from .session import Session

KelvinObjectType = TypeVar("KelvinObjectType", bound="KelvinObject")

logger = logging.getLogger(__name__)


class KelvinObject:
    _class_display_name = "Kelvin Object"
    _kelvin_attrs = ["name", "ucsschool_roles", "udm_properties"]

    def __init__(
        self,
        *,
        name: str = None,
        ucsschool_roles: List[str] = None,
        udm_properties: Dict[str, Any] = None,
        dn: str = None,
        url: str = None,
        session: Session = None,
        language: str = None,
        **kwargs,
    ):
        self.name = name
        self.ucsschool_roles = ucsschool_roles
        self.udm_properties = udm_properties or {}
        self.dn = dn
        self.url = url
        self.session = session
        if language:
            self.session.language = language
        self._resource_class = KelvinResource
        self._fresh = True
        self._deleted = False
        self._old_attrs = {}
        self._update_old_attrs()

    def __repr__(self):
        req_attrs_vals = [f"{attr!r}={value!r}" for attr, value in self._required_get_attrs.items()]
        if hasattr(self, "dn") and self.dn:
            req_attrs_vals.append(f"dn={self.dn!r}")
        return f"{self.__class__.__name__}({', '.join(req_attrs_vals)})"

    async def reload(self) -> KelvinObjectType:
        """
        Reload properties of object from the Kelvin API.

        :raises ucsschool.kelvin.client.NoObject: if the object cannot be found
        :return: self
        """
        if self._deleted:
            logger.warning(
                "[%s] %s %s has been deleted! Trying to load it anyway...",
                self.session.request_id[:10],
                self._class_display_name,
                self,
            )
        obj = await self._resource_class(session=self.session).get(**self._required_get_attrs)
        for k, v in obj.as_dict().items():
            setattr(self, k, v)
        self._update_old_attrs()
        self._fresh = True
        return self

    async def save(self) -> KelvinObjectType:
        if self._deleted:
            raise RuntimeError(f"{self} has been deleted.")
        if not all(self._required_save_attrs.values()):
            raise AssertionError(
                f"{self.__class__.__name__}.save() requires attribute(s) to be set: "
                f"{', '.join(self._resource_class.Meta.required_save_attrs)}."
            )
        if not self._fresh:
            logger.debug(
                "[%s] Saving possibly stale Kelvin object instance.",
                self.session.request_id[:10],
            )
        data = self._to_kelvin_request_data()
        # assumption: if self.url was set, the object exists in the Kelvin API
        # so if it's not set, we'll try to create the object
        if not self.url:
            resp_json = await self.session.post(
                url=self._resource_class(session=self.session).collection_url,
                json=data,
                timeout=30.0,
            )
            resp_obj = self._from_kelvin_response(resp_json)
            for k, v in resp_obj.as_dict().items():
                setattr(self, k, v)
            self._fresh = False
            return self
        # self.url was set -> modify object
        # TODO: or creation failed and this is the fall-back
        resp_json = await self.session.put(url=self.url, json=data)
        resp_obj = self._from_kelvin_response(resp_json)
        for k, v in resp_obj.as_dict().items():
            setattr(self, k, v)
        self._fresh = False
        return self

    async def delete(self) -> None:
        if self._deleted:
            logger.warning("[%s] %s has already been deleted.", self.session.request_id[:10], self)
            return
        if not self.url:
            raise RuntimeError("Attribute 'url' unset. Run 'reload()' before 'delete()'.")
        await self.session.delete(self.url)
        self._deleted = True

    def as_dict(self) -> Dict[str, Any]:
        attrs = self._kelvin_attrs + ["dn", "url"]
        _dict_repr = copy.deepcopy(dict((attr, getattr(self, attr)) for attr in attrs))
        return _dict_repr

    @classmethod
    def _from_kelvin_response(cls, response: Dict[str, Any]) -> KelvinObjectType:
        try:
            # school url to school name
            school_url = response["school"]
            tmp = unquote(school_url.rsplit("/", 1)[-1])
            response["school"] = tmp.split("?")[0]
        except (IndexError, KeyError):
            pass
        return cls(**response)

    def _to_kelvin_request_data(self) -> Dict[str, Any]:
        data = self.as_dict()
        if not data["ucsschool_roles"]:
            del data["ucsschool_roles"]
        del data["dn"]
        del data["url"]
        try:
            # school name to school url
            data["school"] = f"{self.session.urls['school']}{data['school']}"
        except KeyError:
            pass
        return data

    @property
    def _required_get_attrs(self) -> Dict[str, Any]:
        return dict(
            (attr, getattr(self, attr)) for attr in self._resource_class.Meta.required_get_attrs
        )

    @property
    def _required_save_attrs(self) -> Dict[str, Any]:
        return dict(
            (attr, getattr(self, attr)) for attr in self._resource_class.Meta.required_save_attrs
        )

    def _update_old_attrs(self):
        self._old_attrs.update(self._required_get_attrs)


class KelvinResource:
    class Meta:
        kelvin_object: KelvinObjectType = KelvinObject
        required_get_attrs: Iterable[str] = ("name",)
        required_head_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ("school",)

    def __init__(self, session: Session, language: str = None):
        self.session = session
        if language:
            self.session.language = language
        self.collection_url = ""
        self.object_url = ""

    async def get(self, **kwargs) -> KelvinObjectType:
        if not all(attr in kwargs for attr in self.Meta.required_get_attrs):
            raise AssertionError(
                f"{self.__class__.__name__}.get() requires argument(s): "
                f"{', '.join(self.Meta.required_get_attrs)}."
            )
        url = self.object_url.format(**kwargs)
        try:
            return await self.get_from_url(url)
        except NoObject as exc:
            raise NoObject(
                f"{self.Meta.kelvin_object._class_display_name} not found at URL {url!r} using "
                f"attributes {kwargs!r}.",
                reason=exc.reason,
                status=exc.status,
                url=url,
            ) from exc

    async def exists(self, **kwargs) -> bool:
        if not all(attr in kwargs for attr in self.Meta.required_head_attrs):
            raise AssertionError(
                f"{self.__class__.__name__}.get() requires argument(s): "
                f"{', '.join(self.Meta.required_get_attrs)}."
            )
        url = self.object_url.format(**kwargs)
        status_code: int = await self.session.head(url)
        if status_code == 200:
            return True
        if status_code == 404:
            return False
        logger.warning(
            "There was a problem with HEAD for %s. Trying with GET instead. Status code: %s",
            url,
            status_code,
        )
        try:
            await self.get(**kwargs)
        except NoObject:
            return False
        return True

    async def get_from_url(self, url: str) -> KelvinObjectType:
        resp_json: Dict[str, Any] = await self.session.get(url)
        obj = self.Meta.kelvin_object._from_kelvin_response(resp_json)
        obj.session = self.session
        return obj

    async def search(self, **kwargs) -> AsyncIterator[KelvinObjectType]:
        self._check_search_attrs(**kwargs)
        # not necessary, but will simplify the query string
        for k in [k for k, v in kwargs.items() if v in ("", "*")]:
            del kwargs[k]
        resp_json: List[Dict[str, Any]] = await self.session.get(self.collection_url, params=kwargs)
        for resp in resp_json:
            obj = self.Meta.kelvin_object._from_kelvin_response(resp)
            obj.session = self.session
            yield obj

    def _check_search_attrs(self, **kwargs) -> None:
        """
        :raises ucsschool.kelvin.client.InvalidRequest: when there is a problem with the kwargs
            for `search()`
        """
        if not all(attr in kwargs for attr in self.Meta.required_search_attrs):
            raise InvalidRequest(
                f"{self.__class__.__name__}.search() requires argument(s): "
                f"{', '.join(self.Meta.required_search_attrs)}."
            )



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import random
from unittest.mock import patch

import pytest
from faker import Faker

from ucsschool.kelvin.client import RoleResource, Session

fake = Faker()


API_VERSION = "v1"
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_ROLE_RESOURCE = f"{URL_BASE}/{API_VERSION}/roles/"
URL_ROLE_COLLECTION = URL_ROLE_RESOURCE
URL_ROLE_OBJECT = f"{URL_ROLE_RESOURCE}{{name}}"


@pytest.mark.asyncio
async def test_search_no_name_arg(compare_kelvin_obj_with_test_data, kelvin_session_kwargs):
    async with Session(**kelvin_session_kwargs) as session:
        objs = [obj async for obj in RoleResource(session=session).search()]

    assert objs, "No roles found."
    assert len(objs) == 5
    for obj in objs:
        assert not hasattr(obj, "dn")
        assert not hasattr(obj, "ucsschool_roles")
    assert {"staff", "student", "teacher", "school_admin", "legal_guardian"} == {
        obj.name for obj in objs
    }


@pytest.mark.asyncio
@pytest.mark.parametrize("role", ["staff", "student", "teacher", "school_admin", "legal_guardian"])
async def test_get_from_url(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, role):
    url = URL_ROLE_OBJECT.format(host=kelvin_session_kwargs["host"], name=role)
    async with Session(**kelvin_session_kwargs) as session:
        obj = await RoleResource(session=session).get_from_url(url)
    assert obj.name == role


@pytest.mark.asyncio
@pytest.mark.parametrize("role", ["staff", "student", "teacher", "school_admin", "legal_guardian"])
async def test_get(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, role):
    async with Session(**kelvin_session_kwargs) as session:
        obj = await RoleResource(session=session).get(name=role)
    assert obj.name == role


@pytest.mark.asyncio
async def test_role_attrs(compare_kelvin_obj_with_test_data, kelvin_session_kwargs):
    role = random.choice(("staff", "student", "teacher", "school_admin", "legal_guardian"))  # noqa: S311
    async with Session(**kelvin_session_kwargs) as session:
        obj = await RoleResource(session=session).get(name=role)
    assert obj.name == role
    assert set(obj._kelvin_attrs) == {"name", "display_name"}
    assert set(obj.as_dict().keys()) == {"name", "display_name", "url"}


@pytest.mark.asyncio
@pytest.mark.parametrize("role", ["staff", "student", "teacher", "school_admin", "legal_guardian"])
async def test_exists(kelvin_session_kwargs, role):
    async with Session(**kelvin_session_kwargs) as session:
        await RoleResource(session=session).exists(name=role)
        with patch("ucsschool.kelvin.client.session.Session.head") as mock:
            f = asyncio.Future()
            f.set_result(True)
            mock.return_value = f
            await RoleResource(session=session).exists(name=role)

            mock.assert_called_once()


@pytest.mark.asyncio
async def test_exists_non_existing(kelvin_session_kwargs):
    async with Session(**kelvin_session_kwargs) as session:
        assert not await RoleResource(session=session).exists(
            name=fake.user_name(),
        )



# Copyright 2026 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import httpx
import pytest
from async_property import async_property

from ucsschool.kelvin.client.exceptions import NoObject, ServerError
from ucsschool.kelvin.client.session import Session

kelvin_session_kwargs_mock = {
    "username": "username",
    "password": "password",
    "host": "localhost",
    "verify": False,
}


class SessionMock:
    @async_property
    async def token(self) -> str:
        return "Token"


def make_async_mock(results):
    if not isinstance(results, list):
        results = [results]
    it = iter(results)

    async def side_effect(*args, **kwargs):
        res = next(it)
        if isinstance(res, Exception):
            raise res
        return res

    return side_effect


@pytest.mark.asyncio
async def test_session_retry_on_502(mocker):
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    # Mock response to return 502 then 200
    mock_response_502 = mocker.Mock(spec=httpx.Response)
    mock_response_502.status_code = 502
    mock_response_502.reason_phrase = "Bad Gateway"
    mock_response_502.json.return_value = {"detail": "Server error"}

    mock_response_200 = mocker.Mock(spec=httpx.Response)
    mock_response_200.status_code = 200
    mock_response_200.json.return_value = {"success": True}

    mock_get = mocker.patch(
        "httpx.AsyncClient.get",
        side_effect=make_async_mock([mock_response_502, mock_response_502, mock_response_200]),
    )
    mock_get.__name__ = "get"

    # Passing retry_wait_fixed via kwargs to keep tests fast
    async with Session(retries=2, **kelvin_session_kwargs_mock) as session:
        session._max_retry_pause = 0.1
        session._min_retry_pause = 0.1
        resp = await session.get("http://example.com/api")
        assert resp == {"success": True}
        # In pytest-mock, when using side_effect, the mock itself is called
        assert mock_get.call_count == 3


@pytest.mark.asyncio
async def test_session_no_retry_on_404(mocker):
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    mock_response_404 = mocker.Mock(spec=httpx.Response)
    mock_response_404.status_code = 404
    mock_response_404.reason_phrase = "Not Found"
    mock_response_404.json.return_value = {"detail": "Not found"}

    mock_get = mocker.patch("httpx.AsyncClient.get", side_effect=make_async_mock(mock_response_404))
    mock_get.__name__ = "get"

    async with Session(retries=2, **kelvin_session_kwargs_mock) as session:
        session._max_retry_pause = 0.1
        session._min_retry_pause = 0.1
        with pytest.raises(NoObject):
            await session.get("http://example.com/api")
        assert mock_get.call_count == 1


@pytest.mark.asyncio
async def test_session_exhaust_retries(mocker):
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    mock_response_502 = mocker.Mock(spec=httpx.Response)
    mock_response_502.status_code = 502
    mock_response_502.reason_phrase = "Bad Gateway"
    mock_response_502.json.return_value = {"detail": "Server error"}

    mock_get = mocker.patch(
        "httpx.AsyncClient.get", side_effect=make_async_mock([mock_response_502, mock_response_502])
    )
    mock_get.__name__ = "get"

    async with Session(retries=1, **kelvin_session_kwargs_mock) as session:
        session._max_retry_pause = 0.1
        session._min_retry_pause = 0.1
        with pytest.raises(ServerError):
            await session.get("http://example.com/api")
        assert mock_get.call_count == 2


@pytest.mark.asyncio
async def test_session_retries_disabled_by_default(mocker):
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    mock_response_502 = mocker.Mock(spec=httpx.Response)
    mock_response_502.status_code = 502
    mock_response_502.reason_phrase = "Bad Gateway"
    mock_response_502.json.return_value = {"detail": "Server error"}

    mock_get = mocker.patch("httpx.AsyncClient.get", side_effect=make_async_mock(mock_response_502))
    mock_get.__name__ = "get"

    async with Session(**kelvin_session_kwargs_mock) as session:
        with pytest.raises(ServerError):
            await session.get("http://example.com/api")
        assert mock_get.call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "exception",
    [
        httpx.RemoteProtocolError("Server disconnected"),
        httpx.NetworkError("Network issue"),
        httpx.ConnectError("Connection refused"),
    ],
    ids=["RemoteProtocolError", "NetworkError", "ConnectError"],
)
async def test_session_retry_on_network_errors(mocker, exception):
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)

    mock_response_200 = mocker.Mock(spec=httpx.Response)
    mock_response_200.status_code = 200
    mock_response_200.json.return_value = {"success": True}

    mock_get = mocker.patch(
        "httpx.AsyncClient.get",
        side_effect=make_async_mock([exception, mock_response_200]),
    )
    mock_get.__name__ = "get"

    async with Session(retries=1, **kelvin_session_kwargs_mock) as session:
        session._max_retry_pause = 0.1
        session._min_retry_pause = 0.1
        resp = await session.get("http://example.com/api")
        assert resp == {"success": True}
        assert mock_get.call_count == 2



#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import logging
from typing import Any, Dict, Iterable, List, Type

from .base import KelvinObject, KelvinResource
from .session import Session

logger = logging.getLogger(__name__)


class School(KelvinObject):
    _class_display_name = "School"
    _kelvin_attrs = KelvinObject._kelvin_attrs + [
        "display_name",
        "educational_servers",
        "administrative_servers",
        "class_share_file_server",
        "home_share_file_server",
    ]

    def __init__(
        self,
        name: str,
        *,
        display_name: str = None,
        educational_servers: List[str] = None,
        administrative_servers: List[str] = None,
        class_share_file_server: str = None,
        home_share_file_server: str = None,
        ucsschool_roles: List[str] = None,
        udm_properties: Dict[str, Any] = None,
        dn: str = None,
        url: str = None,
        session: Session = None,
        language: str = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            ucsschool_roles=ucsschool_roles,
            udm_properties=udm_properties,
            dn=dn,
            url=url,
            session=session,
            language=language,
        )
        self.display_name = display_name
        self.educational_servers = educational_servers
        self.administrative_servers = administrative_servers
        self.class_share_file_server = class_share_file_server
        self.home_share_file_server = home_share_file_server
        self._resource_class = SchoolResource

    async def save(self) -> "SchoolResource":
        if self.url:
            raise NotImplementedError(
                "Modification of school objects has not yet been implemented "
                "in the Kelvin REST API."
            )
        return await super().save()

    async def delete(self) -> None:
        raise NotImplementedError(
            "Deleting school objects has not yet been implemented in the Kelvin REST API."
        )

    def _to_kelvin_request_data(self) -> Dict[str, Any]:
        data = super()._to_kelvin_request_data()
        # passing None will produce a validation error,
        # OK is not passing it or passing an empty list
        if data["educational_servers"] is None:
            del data["educational_servers"]
        if data["administrative_servers"] is None:
            del data["administrative_servers"]
        return data


class SchoolResource(KelvinResource):
    class Meta:
        kelvin_object: Type[KelvinObject] = School
        required_get_attrs: Iterable[str] = ("name",)
        required_head_attrs: Iterable[str] = ("name",)
        required_save_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ()

    def __init__(self, session: Session, language: str = None):
        super().__init__(session=session, language=language)
        self.collection_url = self.session.urls["school"]
        self.object_url = f"{self.session.urls['school']}{{name}}"



from pathlib import Path

import lazy_object_proxy

try:
    from importlib import metadata
except ImportError:  # pragma: no cover
    # for python < 3.8
    import importlib_metadata as metadata

from .base import KelvinObject, KelvinResource
from .exceptions import (
    InvalidRequest,
    InvalidToken,
    KelvinClientError,
    NoObject,
    ServerError,
)
from .role import Role, RoleResource
from .school import School, SchoolResource
from .school_class import SchoolClass, SchoolClassResource
from .session import Session
from .user import PasswordsHashes, User, UserResource
from .workgroup import WorkGroup, WorkGroupResource

__all__ = [
    "KelvinObject",
    "KelvinResource",
    "InvalidRequest",
    "InvalidToken",
    "KelvinClientError",
    "NoObject",
    "PasswordsHashes",
    "ServerError",
    "School",
    "SchoolResource",
    "SchoolClass",
    "SchoolClassResource",
    "Session",
    "Role",
    "RoleResource",
    "User",
    "UserResource",
    "WorkGroup",
    "WorkGroupResource",
]


def _app_version() -> str:
    try:
        return metadata.version("kelvin-rest-api-client")
    except (metadata.PackageNotFoundError, AttributeError):  # pragma: no cover
        # not yet installed (running tests prior to installation)
        with (Path(__file__).parent.parent.parent.parent / "VERSION.txt").open("r") as fp:
            return fp.read().strip()


__author__ = """Daniel Troeder"""
__email__ = "troeder@univention.de"
__version__: str = lazy_object_proxy.Proxy(_app_version)
